"""
Numpydoc docstring tools.
"""
from __future__ import annotations

import re
from textwrap import dedent

#: List of support docstring sections. The "Fields" section is unique to
#: Eradiate.
NUMPYDOC_SECTION_TITLES = [
    "Parameters",
    "Fields",  # Special section used for classes only
    "Attributes",
    "Returns",
    "Yields",
    "Receives",
    "Other Parameters",
    "Raises",
    "Warns",
    "Warnings",
    "See Also",
    "Notes",
    "References",
    "Examples",
]


def parse_doc(doc: str) -> dict[str, str]:
    """
    Parse a docstring formatted with the Numpydoc style.

    Parameters
    ----------
    doc : str
        Docstring to parse.

    Returns
    -------
    sections : dict
        A dictionary mapping section names to their content. Special
        ``_short_summary`` and ``_extended_summary`` keys are used for
        summary contents.
    """
    doc = dedent(doc.lstrip("\n")).rstrip()

    # We process only sections mentioned in the Numpy doc style
    # Detect existing sections
    section_pattern = {}

    for section_title in NUMPYDOC_SECTION_TITLES:
        full_section_title = f"{section_title}\n{'-' * len(section_title)}\n"
        if re.findall(full_section_title, doc):
            section_pattern[section_title] = full_section_title

    # Collect section content
    section_contents = (
        re.split("|".join(section_pattern.values()), doc) if section_pattern else [doc]
    )

    sections = {}
    if len(section_contents) == len(section_pattern) + 1:
        summary = section_contents.pop(0).strip().split("\n\n")
        sections["_short_summary"] = summary.pop(0)
        if summary:
            sections["_extended_summary"] = "\n\n".join(summary)

    for title, content in zip(section_pattern.keys(), section_contents):
        sections[title] = content.strip()

    return sections


def format_doc(sections: dict[str, str]) -> str:
    """
    Assemble docstring sections and format them according to the Numpydoc style.

    Parameters
    ----------
    sections : dict
        A dictionary mapping section names to contents. Section ordering
        follows the Nompydoc style guide.

    Returns
    -------
    docstring : str
        Formatted docstring.
    """

    # Generate section full text
    section_fulltexts = []

    for section_title in ["_short_summary", "_deprecation", "_extended_summary"]:
        if section_title in sections:
            section_fulltexts.append(sections.pop(section_title) + "\n")

    for section_title in NUMPYDOC_SECTION_TITLES:
        try:
            section_content = sections.pop(section_title)
        except KeyError:
            continue

        section_fulltexts.append(
            f"{section_title}\n{'-' * len(section_title)}\n{section_content}\n"
        )

    # Assemble the final docstring
    doc = "\n".join(section_fulltexts) + "\n"
    return doc



import joseki
import numpy as np
import pytest

from .data import make_particle_dataset
from .. import data
from .. import unit_registry as ureg


@pytest.fixture
def absorption_database_error_handler_config():
    """
    Error handler configuration for absorption coefficient interpolation.

    Notes
    -----
    This configuration is chosen to ignore all interpolation issues (except
    bounds error along the mole fraction dimension) because warnings are
    captured by pytest which will raise.
    Ignoring the bounds on pressure and temperature is safe because
    out-of-bounds values usually correspond to locations in the atmosphere
    that are so high that the contribution to the absorption coefficient
    are negligible at these heights.
    The bounds error for the 'x' (mole fraction) coordinate is considered
    fatal.
    """
    return {
        "p": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        "t": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        "x": {"missing": "ignore", "scalar": "ignore", "bounds": "raise"},
    }


@pytest.fixture
def atmosphere_us_standard_ckd(
    thermoprops_us_standard, absorption_database_error_handler_config
):
    return {
        "type": "molecular",
        "thermoprops": thermoprops_us_standard,
        "absorption_data": "monotropa",
        "error_handler_config": absorption_database_error_handler_config,
    }


@pytest.fixture
def atmosphere_us_standard_mono(
    thermoprops_us_standard, absorption_database_error_handler_config
):
    """
    AFGL (1986) U.S. Standard atmosphere with monochromatic absorption data.

    Notes
    -----
    Molecules included are H2O, CO2, O3, N2O, CO, CH4, O2.
    Specified absorption data covers the wavelength range [250, 3125] nm.
    Altitude grid is regular with a 1 km step, from 0 to 120 km.
    """
    return {
        "type": "molecular",
        "thermoprops": thermoprops_us_standard,
        "absorption_data": "komodo",
        "error_handler_config": absorption_database_error_handler_config,
    }


@pytest.fixture
def atmosphere_cams_lybia4_ckd(absorption_database_error_handler_config):
    """
    CAMS Lybia4 atmosphere with CKD absorption data.

    Notes
    -----
    Molecules included are H2O, CO2, O3, N2O, CO, CH4, O2, NO2, NO, SO2.
    Specified absorption data covers the CKD band associated with wavenumber
    interval [18100, 18200] cm^-1, i.e. the wavelenght range
    [549.45, 552.48] nm.
    Altitude grid is regular with a 1 km step, from 0 to 120 km.
    """
    thermoprops = data.load_dataset("tests/thermoprops/cams_lybia4_2005-04-01.nc")

    return {
        "type": "molecular",
        "thermoprops": thermoprops,
        "absorption_data": "monotropa",
        "error_handler_config": absorption_database_error_handler_config,
    }


@pytest.fixture
def ert_seed_state():
    from eradiate.rng import SeedState

    return SeedState(0)


@pytest.fixture
def particle_dataset_absorbing_only():
    """Absorbing only particles radiative properties data set fixture."""
    mu = np.linspace(-1.0, 1.0) * ureg.dimensionless
    w = np.linspace(279.0, 2401.0) * ureg.nm
    arrays = [np.ones_like(mu) / ureg.steradian for _ in w]
    phase = np.stack(arrays, axis=0).reshape(w.size, mu.size, 1, 1)
    albedo = np.zeros_like(w) * ureg.dimensionless
    sigma_t = np.ones_like(w) / ureg.km
    return make_particle_dataset(
        albedo=albedo, sigma_t=sigma_t, phase=phase, mu=mu, w=w
    )


@pytest.fixture
def particle_dataset_scattering_only():
    """Scattering only particles radiative properties data set fixture."""
    mu = np.linspace(-1.0, 1.0) * ureg.dimensionless
    w = np.linspace(279.0, 2401.0) * ureg.nm
    arrays = [np.ones_like(mu) / ureg.steradian for _ in w]
    phase = np.stack(arrays, axis=0).reshape(w.size, mu.size, 1, 1)
    albedo = np.ones_like(w) * ureg.dimensionless
    sigma_t = np.ones_like(w) / ureg.km
    return make_particle_dataset(
        albedo=albedo, sigma_t=sigma_t, phase=phase, mu=mu, w=w
    )


@pytest.fixture
def particle_dataset_test():
    """
    Particles radiative properties data set fixture.

    Particle radiative properties are spectrally constant with values specified
    below:
        * the phase function is isotropic
        * the extinction coefficient is 1 / km
        * the albedo is 0.8
    """
    mu = np.linspace(-1.0, 1.0) * ureg.dimensionless
    w = np.linspace(279.0, 2401.0) * ureg.nm
    arrays = [np.ones_like(mu) / ureg.steradian for _ in w]
    phase = np.stack(arrays, axis=0).reshape(w.size, mu.size, 1, 1)
    albedo = 0.8 * np.ones_like(w) * ureg.dimensionless
    sigma_t = np.ones_like(w) / ureg.km
    return make_particle_dataset(
        albedo=albedo, sigma_t=sigma_t, phase=phase, mu=mu, w=w
    )


@pytest.fixture(scope="session")
def session_timestamp():
    from datetime import datetime

    return datetime.now()


@pytest.fixture(scope="session")
def thermoprops_us_standard():
    yield joseki.make(
        identifier="afgl_1986-us_standard",
        z=np.linspace(0.0, 120.0, 121) * ureg.km,
        additional_molecules=False,
    )



"""
Particle distributions.

Particle distributions define how the particle number fraction varies with
altitude. The particle layer is split into a number of divisions
(sub-layers) wherein the particle number fraction is evaluated.

Notes
-----
Particle distributions are not normalized. The parent caller is responsible
for normalizing returned values.
"""

import typing as t
from abc import ABC, abstractmethod

import attrs
import numpy as np
import scipy.interpolate

from ..._factory import Factory
from ...attrs import documented, parse_docs

particle_distribution_factory = Factory()
particle_distribution_factory.register_lazy_batch(
    [
        ("UniformParticleDistribution", "uniform", {}),
        ("ExponentialParticleDistribution", "exponential", {}),
        ("GaussianParticleDistribution", "gaussian", {}),
        ("ArrayParticleDistribution", "array", {}),
        ("InterpolatorParticleDistribution", "interpolator", {}),
    ],
    cls_prefix="eradiate.scenes.atmosphere._particle_dist",
)


@parse_docs
@attrs.define
class ParticleDistribution(ABC):
    """
    Abstract base class for particle distributions used to define particle
    layers.

    In practice, particle distributions are callables with the signature
    ``f(x: np.typing.ArrayLike) -> np.ndarray`` and are evaluated over the
    interval [0, 1].
    """

    @abstractmethod
    def __call__(self, x: np.typing.ArrayLike) -> np.ndarray:
        pass


@parse_docs
@attrs.define
class UniformParticleDistribution(ParticleDistribution):
    r"""
    Uniform particle distribution. Returns values given by the uniform PDF

    .. math::
       f : x \mapsto \left\{
           \begin{array}{ll}
               \frac{1}{b - a} & \mathrm{if} \ x \in [a, b] \\
               0 & \mathrm{otherwise}
           \end{array}
       \right.

    where :math:`a = \mathtt{bounds[0]}` and :math:`b = \mathtt{bounds[1]}`.
    """

    bounds: np.ndarray = documented(
        attrs.field(
            default=[0.0, 1.0], converter=lambda x: np.atleast_1d(np.squeeze(x))
        ),
        type="array",
        init_type="array-like, optional",
        default="[0, 1]",
        doc="Bounds of the distribution's interval.",
    )

    @bounds.validator
    def _bounds_validator(self, attribute, value):
        if len(value) != 2:
            raise ValueError(
                f"while validating '{attribute.name}': passed array must have "
                "exactly 2 elements"
            )

        if value[1] <= value[0]:
            raise ValueError(
                f"while validating '{attribute.name}': bounds must be sorted in "
                "ascending order "
            )

    def __call__(self, x: np.typing.ArrayLike) -> np.ndarray:
        return np.where(
            np.logical_or(x < self.bounds[0], x > self.bounds[1]),
            np.zeros_like(x),
            np.full_like(x, 1.0 / (self.bounds[1] - self.bounds[0])),
        )


@parse_docs
@attrs.define(init=False)
class ExponentialParticleDistribution(ParticleDistribution):
    r"""
    Exponential particle distribution. Returns values given by the normalized
    exponential PDF

    .. math::
       f : x \mapsto \dfrac{\lambda e^{-\lambda x}}{1 - e^{-\lambda}}

    where :math:`\lambda = \mathtt{rate}`.

    Parameters
    ----------
    rate : float, optional, default: 5.0
        Decay rate :math:`\lambda`. The default value ensures a 99.3% decay on
        the [0, 1] interval. Mutually exclusive with `scale`.

    scale : float, optional
        Scale :math:`\beta = 1 / \lambda`. Mutually exclusive with `rate`.

    Fields
    ------
    rate : float
        Decay rate :math:`\lambda`.
    """

    rate: float = attrs.field(default=5.0, converter=float)

    @rate.validator
    def _rate_validator(self, attribute, value):
        if value <= 0:
            raise ValueError(
                f"while validating {attribute.name}: rate must be strictly positive"
            )

    def __init__(self, rate: float = None, scale: float = None):
        if rate is None and scale is None:
            self.__attrs_init__()
        elif scale is None and rate is not None:
            self.__attrs_init__(rate)
        elif scale is not None and rate is None:
            self.__attrs_init__(1.0 / scale)
        else:
            raise ValueError(
                "while initialising ExponentialParticleDistribution: "
                "setting both rate and scale is not allowed"
            )

    def __call__(self, x: np.typing.ArrayLike) -> np.ndarray:
        return np.where(
            np.logical_or(x < 0.0, x > 1.0),
            np.zeros_like(x),
            np.exp(-x * self.rate) * self.rate / (1.0 - np.exp(-self.rate)),
        )


@parse_docs
@attrs.define
class GaussianParticleDistribution(ParticleDistribution):
    r"""
    Gaussian particle distribution. Returns values given by the Gaussian
    PDF

    .. math::
       f : x \mapsto \frac{1}{2 \pi \cdot \sigma}
           \exp \left[
             -\frac{1}{2}
             \left( \frac{x - \mu}{\sigma} \right)^2
           \right]

    where :math:`\mu = \mathtt{mean}` and :math:`\sigma = \mathtt{std}`.
    """

    mean: float = documented(
        attrs.field(default=0.5, converter=float),
        type="float",
        init_type="float, optional",
        default="0.5",
        doc="Mean of the Gaussian PDF. The default value places the mean in "
        "the middle of the particle layer (at :math:`x = 0.5`).",
    )

    std: float = documented(
        attrs.field(default=0.5 / 3, converter=float),
        type="float",
        init_type="float, optional",
        default="1/6",
        doc="Standard deviation of the Gaussian PDF. The default value is "
        "such that the integral of the Gaussian PDF over the "
        r":math:`[\mu - 0.5, \mu + 0.5]` interval is about 99.7% (3σ).",
    )

    def __call__(self, x: np.typing.ArrayLike) -> np.ndarray:
        return np.where(
            np.logical_or(x < 0.0, x > 1.0),
            np.zeros_like(x),
            np.exp(-0.5 * np.square((x - self.mean) / self.std))
            / (self.std * np.sqrt(2.0 * np.pi)),
        )


@parse_docs
@attrs.define
class ArrayParticleDistribution(ParticleDistribution):
    """
    Particle distribution specified by an array of values.
    """

    values: np.typing.ArrayLike = documented(
        attrs.field(converter=np.array, kw_only=True),
        type="ndarray",
        init_type="array-like",
        doc="An array of particle fraction values.",
    )

    @values.validator
    def _values_validator(self, attribute, value):
        if value.ndim != 1:
            raise ValueError(
                f"while validating {attribute.name}: only 1D arrays are allowed"
            )

        if len(value) < 2:
            raise ValueError(
                f"while validating {attribute.name}: array must have at least 2 "
                "elements"
            )

    coords: np.ndarray = documented(
        attrs.field(
            default=attrs.Factory(
                lambda x: np.arange(
                    0.5 / len(x.values),
                    1,
                    1 / len(x.values),
                ),
                takes_self=True,
            ),
            converter=np.array,
        ),
        type="ndarray",
        init_type="array-like, optional",
        doc="Coordinates to which passed values are mapped. This array must "
        "have the same shape as ``values``. The default value positions values "
        "at the centers of a regular grid with nodes defined "
        "by :code:`np.linspace(0, 1, len(values))`.",
    )

    @coords.validator
    def _coords_validator(self, attribute, value):
        if value.shape != self.values.shape:
            raise ValueError(
                f"while validating '{attribute.name}': coordinate and value "
                "array shapes must be the same"
            )

    method: str = documented(
        attrs.field(
            default="linear",
            converter=str,
            validator=attrs.validators.in_(
                {
                    "linear",
                    "nearest",
                    "nearest-up",
                    "zero",
                    "slinear",
                    "quadratic",
                    "cubic",
                    "previous",
                    "next",
                }
            ),
        ),
        type="str",
        init_type='{ "linear", "nearest", "nearest-up", "zero", "slinear", '
        '"quadratic", "cubic", "previous", "next" }',
        default='"linear"',
        doc="Interpolation method. See :class:`scipy.interpolate.interp1d` "
        "(*kind*) for more information.",
    )

    extrapolate: str = documented(
        attrs.field(
            default="zero",
            converter=str,
            validator=attrs.validators.in_({"zero", "nearest", "method", "nan"}),
        ),
        type="str",
        init_type='{ "zero", "nearest", "method", "nan" }',
        default='"zero"',
        doc="Extrapolation method used when evaluation is requested outside of "
        r":math:`[\mathtt{coords[0]}, \mathtt{coords[-1]}]`. "
        "See :class:`scipy.interpolate.interp1d` (*fill_value*) for more "
        "information. Settings map as follows:\n"
        "\n"
        ".. list-table::\n\n"
        "   * - ``ArrayParticleDistribution`` (`extrapolate`)\n"
        "     - ``interp1d`` (`fill_value`)\n"
        '   * - ``"zero"``\n'
        "     - ``0.0``\n"
        '   * - ``"nearest"``\n'
        "     - ``(values[0], values[-1])``\n"
        '   * - ``"method"``\n'
        "     - ``extrapolate``\n"
        '   * - ``"nan"``\n'
        "     - ``np.nan``\n",
    )

    def __call__(self, x: np.typing.ArrayLike) -> np.ndarray:
        if self.extrapolate == "zero":
            fill_value = 0.0
        elif self.extrapolate == "nearest":
            fill_value = (self.values[0], self.values[-1])
        elif self.extrapolate == "method":
            fill_value = "extrapolate"
        else:
            fill_value = np.nan

        # TODO: This is unsafe, values of x outside of [0, 1] should return 0
        f = scipy.interpolate.interp1d(
            self.coords,
            self.values,
            kind=self.method,
            bounds_error=False,
            fill_value=fill_value,
        )
        return f(x)


@parse_docs
@attrs.define
class InterpolatorParticleDistribution(ParticleDistribution):
    """
    A flexible particle distribution which redirects its calls to an
    encapsulated callable.
    """

    interpolator: t.Callable[[np.typing.ArrayLike], np.ndarray] = documented(
        attrs.field(validator=attrs.validators.is_callable, kw_only=True),
        type="callable",
        doc="A callable with signature "
        ":code:`f(x: np.typing.ArrayLike) -> np.ndarray`. Typically, "
        "a :class:`scipy.interpolate.interp1d`.",
    )

    def __call__(self, x: np.typing.ArrayLike) -> np.ndarray:
        # TODO: This is unsafe, values of x outside of [0, 1] should return 0
        return self.interpolator(x)



"""The Eradiate radiative transfer simulation software package."""

from ._version import _version

__version__ = _version  #: Eradiate version string.

# -- Lazy imports ------------------------------------------------------

import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader



from ._canopies import wellington_citrus_orchard as wellington_citrus_orchard
from ._canopy_loader import load_scenario as load_scenario
from ._core import Canopy as Canopy
from ._core import CanopyElement as CanopyElement
from ._core import InstancedCanopyElement as InstancedCanopyElement
from ._core import biosphere_factory as biosphere_factory
from ._discrete import DiscreteCanopy as DiscreteCanopy
from ._leaf_cloud import LeafCloud as LeafCloud
from ._rami_scenarios import RAMIActualCanopies as RAMIActualCanopies
from ._rami_scenarios import (
    RAMIHeterogeneousAbstractCanopies as RAMIHeterogeneousAbstractCanopies,
)
from ._rami_scenarios import (
    RAMIHomogeneousAbstractCanopies as RAMIHomogeneousAbstractCanopies,
)
from ._rami_scenarios import RAMIScenarioVersion as RAMIScenarioVersion
from ._rami_scenarios import load_rami_scenario as load_rami_scenario
from ._tree import AbstractTree as AbstractTree
from ._tree import MeshTree as MeshTree
from ._tree import MeshTreeElement as MeshTreeElement



"""
Atmosphere's radiative profile.
"""

from __future__ import annotations

import attrs
import joseki
import numpy as np
import pint
import xarray as xr
from joseki.profiles.core import interp

from ._absorption import AbsorptionDatabase
from ._core import RadProfile, ZGrid, make_dataset
from .rayleigh import compute_sigma_s_air
from ..attrs import documented, parse_docs
from ..converters import convert_thermoprops
from ..units import to_quantity
from ..units import unit_registry as ureg
from ..util.misc import cache_by_id, summary_repr


@parse_docs
@attrs.define(eq=False)
class AtmosphereRadProfile(RadProfile):
    """
    Atmospheric radiative profile.

    This class provides an interface to generate vertical profiles of
    atmospheric volume radiative properties (also sometimes referred to as
    collision coefficients).

    The atmospheric radiative profile is built from a thermophysical profile,
    which provides the temperature, pressure and species concentrations as a
    function of altitude, and an absorption coefficient database indexed by
    those thermophysical variables.
    """

    absorption_data: AbsorptionDatabase = documented(
        attrs.field(
            factory=AbsorptionDatabase.default,
            converter=AbsorptionDatabase.convert,
            validator=attrs.validators.instance_of(AbsorptionDatabase),
        ),
        doc="Absorption coefficient data. The passed value is pre-processed by "
        ":meth:`.AbsorptionDatabase.convert`.",
        type="AbsorptionDatabase",
        init_type="str or path-like or dict or .AbsorptionDatabase",
        default=":meth:`AbsorptionDatabase.default() <.AbsorptionDatabase.default>`",
    )

    thermoprops: xr.Dataset = documented(
        attrs.field(
            factory=lambda: joseki.make(
                identifier="afgl_1986-us_standard",
                z=np.linspace(0.0, 120.0, 121) * ureg.km,
                additional_molecules=False,
            ),
            converter=convert_thermoprops,
            repr=summary_repr,
        ),
        doc="Atmosphere's thermophysical properties.",
        type="Dataset",
        default=(
            "`joseki.make <https://rayference.github.io/joseki/latest/reference/#src.joseki.core.make>`_"
            ' with ``identifier`` set to "afgl_1986-us_standard" and '
            '``z`` set to "np.linspace(0.0, 120.0, 121) * ureg.km" and '
            '``additional_molecules`` set to "False".'
        ),
    )

    @thermoprops.validator
    def _check_thermoprops(self, attribute, value):
        if not value.joseki.is_valid:
            raise ValueError(
                "Invalid thermophysical properties dataset."  # TODO: explain what is invalid
            )

    has_absorption: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    _zgrid: ZGrid | None = attrs.field(default=None, init=False)

    def __attrs_post_init__(self):
        self.update()

    def update(self) -> None:
        self._zgrid = ZGrid(levels=self.levels)

    @property
    def levels(self) -> pint.Quantity:
        return to_quantity(self.thermoprops.z)

    @property
    def zgrid(self) -> ZGrid:
        # Inherit docstring
        return self._zgrid

    @cache_by_id
    def _thermoprops_interp(self, zgrid: ZGrid) -> xr.Dataset:
        # Interpolate thermophysical profile on specified altitude grid
        # Note: this value is cached so that repeated calls with the same zgrid
        #       won't trigger an unnecessary computation.
        return interp(
            self.thermoprops,
            z_new=zgrid.levels.m * ureg(str(zgrid.levels.units)),
            method={"default": "nearest"},  # TODO: revisit
        )

    def eval_albedo_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Inherit docstring
        sigma_s = self.eval_sigma_s_mono(w, zgrid)
        sigma_t = self.eval_sigma_t_mono(w, zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_albedo_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        sigma_t = self.eval_sigma_t_ckd(w=w, g=g, zgrid=zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_sigma_a_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # NOTE: this method accepts 'w'-arrays and is vectorized as far as
        # each individual absorption dataset is concerned, namely when the
        # wavelengths span multiple datasets we for-loop over them.
        w = np.atleast_1d(w)
        if self.has_absorption:
            thermoprops = self._thermoprops_interp(zgrid)
            values = self.absorption_data.eval_sigma_a_mono(w, thermoprops).transpose(
                "w", "z"
            )
            values = to_quantity(values)
            # We evaluated the
            # project on altitude layers
            return 0.5 * (values[:, 1:] + values[:, :-1]).squeeze()
        else:
            return np.zeros((w.size, zgrid.n_layers)).squeeze() / ureg.km

    def eval_sigma_a_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        w = np.atleast_1d(w)
        if self.has_absorption:
            values = self.absorption_data.eval_sigma_a_ckd(
                w, g, self._thermoprops_interp(zgrid)
            )  # axis order = (w, z)
            values = to_quantity(values)

            # Interpolate on altitude layers
            return 0.5 * (values[:, 1:] + values[:, :-1]).squeeze()
        else:
            return np.zeros((w.size, zgrid.n_layers)).squeeze() / ureg.km

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        if self.has_scattering:
            thermoprops = self._thermoprops_interp(zgrid)
            sigma_s = compute_sigma_s_air(
                wavelength=w,
                number_density=to_quantity(thermoprops.n),
            )
            # project on altitude layers
            return 0.5 * (sigma_s[1:] + sigma_s[:-1])
        else:
            return np.zeros((1, zgrid.n_layers)) / ureg.km

    def eval_sigma_s_ckd(
        self, w: pint.Quantity, g: float, zgrid: ZGrid
    ) -> pint.Quantity:
        return self.eval_sigma_s_mono(w=w, zgrid=zgrid)

    def eval_sigma_t_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        sigma_a = self.eval_sigma_a_mono(w=w, zgrid=zgrid)
        sigma_s = self.eval_sigma_s_mono(w=w, zgrid=zgrid)
        return sigma_a + sigma_s

    def eval_sigma_t_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        sigma_a = self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid)
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        return sigma_a + sigma_s

    def eval_dataset_mono(self, w: pint.Quantity, zgrid: ZGrid) -> xr.Dataset:
        return make_dataset(
            wavelength=w,
            z_level=zgrid.levels,
            z_layer=zgrid.layers,
            sigma_a=self.eval_sigma_a_mono(w, zgrid),
            sigma_s=self.eval_sigma_s_mono(w, zgrid),
        ).squeeze()

    def eval_dataset_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> xr.Dataset:
        return make_dataset(
            wavelength=w,
            z_level=zgrid.levels,
            z_layer=zgrid.layers,
            sigma_a=self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid),
            sigma_s=self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid),
        ).squeeze()



from __future__ import annotations

import attrs
import numpy as np
import pint
import pinttr

from ._core import Measure
from ... import validators
from ...attrs import documented, parse_docs
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


@parse_docs
@attrs.define(eq=False, slots=False)
class RadiancemeterMeasure(Measure):
    """
    Radiance meter measure scene element [``radiancemeter``].

    This measure scene element is a thin wrapper around the ``radiancemeter``
    sensor kernel plugin. It records the incident power per unit area per unit
    solid angle along a certain ray.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    origin: pint.Quantity = documented(
        pinttr.field(
            default=ureg.Quantity([0.0, 0.0, 0.0], ureg.m),
            validator=[validators.has_len(3), pinttr.validators.has_compatible_units],
            units=ucc.deferred("length"),
        ),
        doc="A 3-element vector specifying the position of the radiance meter.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[0, 0, 0] m",
    )

    target: pint.Quantity = documented(
        pinttr.field(
            default=ureg.Quantity([0.0, 0.0, 1.0], ureg.m),
            validator=[validators.has_len(3), pinttr.validators.has_compatible_units],
            units=ucc.deferred("length"),
        ),
        doc="A 3-element vector specifying the location targeted by the sensor.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[0, 0, 1] m",
    )

    @target.validator
    @origin.validator
    def _target_origin_validator(self, attribute, value):
        if np.allclose(self.target, self.origin):
            raise ValueError(
                f"While initializing {attribute}: "
                f"Origin and target must not be equal, "
                f"got target = {self.target}, origin = {self.origin}"
            )

    @property
    def film_resolution(self) -> tuple[int, int]:
        return (1, 1)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        return "radiancemeter"

    @property
    def template(self) -> dict:
        result = super().template

        target = self.target.m_as(uck.get("length"))
        origin = self.origin.m_as(uck.get("length"))
        direction = target - origin
        result["origin"] = origin
        result["direction"] = direction

        return result

    # --------------------------------------------------------------------------
    #                        Post-processing information
    # --------------------------------------------------------------------------

    @property
    def var(self) -> tuple[str, dict]:
        return "radiance", {
            "standard_name": "radiance",
            "long_name": "radiance",
            "units": symbol(uck.get("radiance")),
        }



"""
Post-processing pipeline definitions.
"""

from __future__ import annotations

import numpy as np
import xarray as xr
from hamilton.function_modifiers import (
    ResolveAt,
    config,
    extract_fields,
    parameterize,
    resolve,
    source,
    tag,
    tag_outputs,
    value,
)

from . import logic
from .._mode import modes
from ..scenes.illumination import (
    Illumination,
)
from ..scenes.spectra import (
    Spectrum,
)
from ..spectral.spectral_set import SpectralSet

_MODE_IDS_CKD = set(modes(lambda x: x.is_ckd).keys())


def _parameterize_aggregate_ckd_quad(var_name):
    # Tailored decorator for the aggregate_ckd_quad() task
    cfg = {
        var_name: {"raw_data": source(f"{var_name}_raw")}
    }
    return parameterize(**cfg)


@tag(**{"final": "true", "kind": "data"})
@resolve(
    when=ResolveAt.CONFIG_AVAILABLE,
    decorate_with=_parameterize_aggregate_ckd_quad,
)
def aggregate_ckd_quad(
    mode_id: str, raw_data: xr.DataArray, spectral_set: SpectralSet
) -> xr.DataArray:
    return logic.aggregate_ckd_quad(mode_id, raw_data, spectral_set, False)


def _parameterize_aggregate_ckd_quad_var(var_name):
    # Tailored decorator for the aggregate_ckd_quad_var() task
    cfg = {
        f"{var_name}_var": {"raw_data": source(f"{var_name}_var_raw")},
    }
    return parameterize(**cfg)


@tag(**{"final": "true", "kind": "data"})
@config.when(calculate_variance=True)
@resolve(
    when=ResolveAt.CONFIG_AVAILABLE,
    decorate_with=_parameterize_aggregate_ckd_quad_var,
)
def aggregate_ckd_quad_var(
    mode_id: str, raw_data: xr.DataArray, spectral_set: SpectralSet
) -> xr.DataArray:
    return logic.aggregate_ckd_quad(mode_id, raw_data, spectral_set, True)


def _parameterize_variance(var_name):
    # Tailored decorator for the moment2_to_variance() task
    cfg = {
        f"{var_name}_var_raw": {
            "raw_data": source(f"{var_name}_raw"),
            "raw_m2_data": source(f"{var_name}_m2_raw"),
        }
    }
    return parameterize(**cfg)


@tag(**{"kind": "data"})
@config.when(calculate_variance=True)
@resolve(
    when=ResolveAt.CONFIG_AVAILABLE,
    decorate_with=_parameterize_variance,
)
def moment2_to_variance(
    raw_data: xr.DataArray, raw_m2_data: xr.DataArray, spp: xr.DataArray
) -> xr.DataArray:
    return logic.moment2_to_variance(raw_data, raw_m2_data, spp)


def _parameterize_apply_spectral_response(var_name, measure_distant):
    # Tailored decorator for the _apply_spectral_response() task

    # Add basic measure variable
    cfg = {f"{var_name}_srf": {"spectral_data": source(var_name)}}

    # Possibly add other derived spectral variables
    if var_name == "sector_radiosity":
        cfg["radiosity_srf"] = {"spectral_data": source("radiosity")}

    if measure_distant:
        cfg["irradiance_srf"] = {"spectral_data": source("irradiance")}

    return parameterize(**cfg)


@tag(**{"final": "true", "kind": "data"})
@resolve(
    when=ResolveAt.CONFIG_AVAILABLE, decorate_with=_parameterize_apply_spectral_response
)
@config.when_in(mode_id=_MODE_IDS_CKD)
@config.when(apply_spectral_response=True)
def apply_spectral_response(spectral_data: xr.DataArray, srf: Spectrum) -> xr.DataArray:
    return logic.apply_spectral_response(spectral_data, srf)


@tag(**{"final": "true", "kind": "data"})
@config.when(var_name="sector_radiosity", measure_distant=True)
def albedo(radiosity: xr.DataArray, irradiance: xr.DataArray) -> xr.DataArray:
    return logic.compute_albedo(radiosity, irradiance)


@tag(**{"final": "true", "kind": "data"})
@config.when(
    var_name="sector_radiosity", measure_distant=True, apply_spectral_response=True
)
def albedo_srf(
    radiosity_srf: xr.DataArray, irradiance_srf: xr.DataArray
) -> xr.DataArray:
    return logic.compute_albedo(radiosity_srf, irradiance_srf)


@tag_outputs(
    **{
        "brdf": {"final": "true", "kind": "data"},
        "brf": {"final": "true", "kind": "data"},
    }
)
@config.when(var_name="radiance", measure_distant=True)
@extract_fields({"brdf": xr.DataArray, "brf": xr.DataArray})
def bidirectional_reflectance(radiance: xr.DataArray, irradiance: xr.DataArray) -> dict:
    result = logic.compute_bidirectional_reflectance(radiance, irradiance)
    return {k: result[k] for k in result.keys()}


@tag_outputs(
    **{
        "brdf_srf": {"final": "true", "kind": "data"},
        "brf_srf": {"final": "true", "kind": "data"},
    }
)
@config.when_in(mode_id=_MODE_IDS_CKD)
@config.when(var_name="radiance", measure_distant=True, apply_spectral_response=True)
@extract_fields({"brdf_srf": xr.DataArray, "brf_srf": xr.DataArray})
def bidirectional_reflectance_srf(
    radiance_srf: xr.DataArray, irradiance_srf: xr.DataArray
) -> dict:
    result = logic.compute_bidirectional_reflectance(radiance_srf, irradiance_srf)
    return {"brdf_srf": result["brdf"], "brf_srf": result["brf"]}


@tag_outputs(
    **{
        "irradiance": {"final": "true", "kind": "data"},
        "solar_angles": {"final": "true", "kind": "coord"},
    }
)
@extract_fields({"irradiance": xr.DataArray, "solar_angles": xr.Dataset})
def extract_irradiance(
    mode_id: str, illumination: Illumination, spectral_set: SpectralSet
) -> dict:
    return logic.extract_irradiance(mode_id, illumination, spectral_set)


@resolve(
    when=ResolveAt.CONFIG_AVAILABLE,
    decorate_with=lambda var_name: tag_outputs(
        **{
            "spp": {"kind": "data"},
            f"{var_name}_raw": {"kind": "data"},
            f"{var_name}_m2_raw": {"kind": "data"},
        }
    ),
)
@resolve(
    when=ResolveAt.CONFIG_AVAILABLE,
    decorate_with=lambda var_name: extract_fields(
        {
            # Requested sample count for the corresponding Mitsuba render pass
            "spp": xr.DataArray,
            # Raw pixel weights (equal to collected sampled count with a box filter)
            "weights_raw": xr.DataArray,
            # Main film pixel variable being processed
            f"{var_name}_raw": xr.DataArray,
            # Wavelength list, source from bitmap data
            # "wavelengths": xr.DataArray,
            # 2nd moment of the main film
            f"{var_name}_m2_raw": xr.DataArray,
        }
    ),
)
def gather_bitmaps(
    mode_id: str,
    var_name: str,
    var_metadata: dict,
    calculate_variance: bool,
    bitmaps: dict,
    viewing_angles: xr.Dataset,
    solar_angles: xr.Dataset,
) -> dict:
    return logic.gather_bitmaps(
        mode_id,
        var_name,
        var_metadata,
        calculate_variance,
        bitmaps,
        viewing_angles,
        solar_angles,
    )


@tag(**{"final": "true", "kind": "data"})
@config.when(var_name="sector_radiosity")
def radiosity(sector_radiosity: xr.DataArray) -> xr.DataArray:
    return logic.radiosity(sector_radiosity)


@config.when(apply_spectral_response=True)
def spectral_response(srf: Spectrum) -> xr.DataArray:
    return logic.spectral_response(srf)


@tag(**{"final": "true", "kind": "coord"})
@config.when(add_viewing_angles=True)
def viewing_angles(angles: np.ndarray) -> xr.Dataset:
    return logic.viewing_angles(angles)



"""
Components related with pseudo-random number generation.

Inspired by `SeedBank <https://github.com/lenskit/seedbank>`_.
"""
from __future__ import annotations

import attrs
import numpy as np
import numpy.random


@attrs.define
class SeedState:
    """
    Manage a root seed and facilities to derive seeds.
    """

    _seed: np.random.SeedSequence | None = attrs.field(
        default=None,
        converter=lambda x: x
        if isinstance(x, np.random.SeedSequence)
        else np.random.SeedSequence(x),
    )

    def reset(self, seed=None):
        """
        Reset the seed state.

        Parameters
        ----------
        seed : int or numpy.random.SeedSequence, optional
            Value used to initialize the internal seed sequence. If unset, the
            current seed sequence is reused, with its children spawned member
            reset.
        """
        if seed is not None:
            self._seed = (
                seed
                if isinstance(seed, np.random.SeedSequence)
                else np.random.SeedSequence(seed)
            )
        else:
            self._seed = np.random.SeedSequence(entropy=self._seed.entropy)

    def next(self, n: int = 1) -> np.ndarray:
        """
        Get the next *n* seed values.

        Parameters
        ----------
        n : int
            Number of seed values to generate.

        Returns
        -------
        ndarray
            Generated RNG seeds.
        """
        result = self._seed.spawn(1)[0].generate_state(n)
        return result

    def numpy_default_rng(self) -> numpy.random.Generator:
        """
        Return a default Numpy RNG initialized with a generated seed.

        Returns
        -------
        numpy.random.Generator
            Initialised RNG.
        """
        seed = self.next(1)[0]
        return np.random.default_rng(seed=seed)


#: Deterministic root seed state (see :class:`.SeedState`).
root_seed_state = SeedState(0)



import typing as t

import numpy as np
import pandas as pd

from ._core import InstancedCanopyElement
from ._tree import MeshTree
from ..spectra import InterpolatedSpectrum
from ... import data
from ...units import unit_registry as ureg


def wellington_citrus_orchard(
    padding=0, properties: t.Literal["rami", "hyperspectral"] = "rami"
) -> dict:
    """
    Generate a base experiment dictionary based on the Wellington Citrus Orchard
    scene from the RAMI benchmark series.

    Parameters
    ----------
    padding : int, optional, default: 0
        Amount of padding applied to the generated canopy.

    properties : {"rami", "hyperspectral"}, optional, default: "rami"
        If set to ``"rami"``, the radiative properties used for the RAMI
        benchmark series is used. If set to ``"hyperspectral"``, leaf and soil
        reflectance are replaced with spectral datasets.

    Returns
    -------
    dict
        A dictionary suitable for keyword argument specification for the
        :class:`.CanopyExperiment` and :class:`CanopyAtmosphereExperiment`
        constructors. It contains the surface and canopy specifications.

    Warnings
    --------
    * This is an experimental feature, please use with caution and report issues.
    * The underlying data is sourced from the scene file archived distributed by
      the DART team and post-processed to accommodate Eradiate-specific frame
      conventions and formats. We do not recommend using it to run RAMI
      benchmark cases.
    """
    spectrum_table = pd.read_csv(
        data.data_store.fetch("trees/citrus_sinensis/spectral.txt"),
        sep=r"\s+",
        comment="#",
        index_col=0,
    )

    # Collect wavelengths
    w = (
        0.5
        * (spectrum_table.loc["WLMIN"].values + spectrum_table.loc["WLMAX"].values)
        * ureg.nm
    )  # Central wavelength for each band

    # Collect trunk reflectance
    trunk_reflectance = InterpolatedSpectrum(
        quantity="reflectance",
        wavelengths=w,
        values=spectrum_table.loc["CTR[1-10]_WOOD_R"].values,
    )

    if properties == "rami":
        # Collect ground reflectance
        ground_reflectance = InterpolatedSpectrum(
            quantity="reflectance",
            wavelengths=w,
            values=spectrum_table.loc["BGROUND_REFL"].values,
        )

        # Collect foliage reflectance/transmittance and define individual trees
        citrus_sinensis = []

        for i in range(1, 11):
            transmittance = InterpolatedSpectrum(
                quantity="transmittance",
                wavelengths=w,
                values=spectrum_table.loc[f"CTR{i}_FOLI_TRAN"].values,
            )
            reflectance = InterpolatedSpectrum(
                quantity="reflectance",
                wavelengths=w,
                values=spectrum_table.loc[f"CTR{i}_FOLI_REFL"].values,
            )
            citrus_sinensis.append(
                MeshTree(
                    id=f"cisi{i}",
                    mesh_tree_elements=[
                        {
                            "id": f"trunk{i}",
                            "mesh_filename": data.data_store.fetch(
                                f"trees/citrus_sinensis/CISI{i}/trunk.ply"
                            ),
                            "reflectance": trunk_reflectance,
                        },
                        {
                            "id": f"foliage{i}",
                            "mesh_filename": data.data_store.fetch(
                                f"trees/citrus_sinensis/CISI{i}/foliage.ply"
                            ),
                            "transmittance": transmittance,
                            "reflectance": reflectance,
                        },
                    ],
                )
            )

    elif properties == "hyperspectral":
        # Collect ground reflectance
        ground_ds = data.open_dataset("spectra/reflectance/lambertian_soil.nc").sel(
            brightness="dark", drop=True
        )
        ground_reflectance = InterpolatedSpectrum(
            quantity="reflectance",
            wavelengths=ureg.Quantity(
                ground_ds.w.values,
                ground_ds.w.units,
            ),
            values=ground_ds["reflectance"].values,
        )

        # Collect foliage reflectance/transmittance and define individual trees
        citrus_sinensis = []
        foliage_ds = data.open_dataset(
            "spectra/reflectance/bilambertian_leaf_cellulose.nc"
        )
        w = ureg.Quantity(foliage_ds.w.values, foliage_ds.w.units)
        transmittance = InterpolatedSpectrum(
            quantity="transmittance",
            wavelengths=w,
            values=foliage_ds.transmittance.values,
        )
        reflectance = InterpolatedSpectrum(
            quantity="reflectance",
            wavelengths=w,
            values=foliage_ds.reflectance.values,
        )

        for i in range(1, 11):
            citrus_sinensis.append(
                MeshTree(
                    id=f"cisi{i}",
                    mesh_tree_elements=[
                        {
                            "id": f"trunk{i}",
                            "mesh_filename": data.data_store.fetch(
                                f"trees/citrus_sinensis/CISI{i}/trunk.ply"
                            ),
                            "reflectance": trunk_reflectance,
                        },
                        {
                            "id": f"foliage{i}",
                            "mesh_filename": data.data_store.fetch(
                                f"trees/citrus_sinensis/CISI{i}/foliage.ply"
                            ),
                            "transmittance": transmittance,
                            "reflectance": reflectance,
                        },
                    ],
                )
            )

    else:
        raise ValueError(f"unknown spectral property specification '{properties}'")

    # Load instance positions
    citrus_sinensis_locations = []

    for i in range(1, 11):
        loc = np.genfromtxt(
            data.data_store.fetch(f"trees/citrus_sinensis/CISI{i}/location.txt")
        )
        zeros = np.zeros((len(loc), 1))
        citrus_sinensis_locations.append(np.concatenate((loc, zeros), axis=1))

    instanced_elements = []

    for i, tree, locations in zip(
        range(1, 11), citrus_sinensis, citrus_sinensis_locations
    ):
        instanced_elements.append(
            InstancedCanopyElement(
                canopy_element=tree,
                instance_positions=locations,
            )
        )

    return {
        "surface": {"type": "lambertian", "reflectance": ground_reflectance},
        "canopy": {
            "type": "discrete_canopy",
            "instanced_canopy_elements": instanced_elements,
            "size": [100, 100, 10],
            "padding": padding,
        },
    }



import mitsuba as mi
import typer
from rich.console import Console

app = typer.Typer()


@app.command()
def main():
    """
    Display information useful for debugging.
    """
    import eradiate.util.sys_info

    console = Console(color_system=None)
    error_console = Console(stderr=True, color_system=None)
    sys_info = eradiate.util.sys_info.show()

    def section(title, newline=True):
        if newline:
            console.print()
        console.rule("── " + title, align="left")
        console.print()

    def message(text):
        console.print(text)

    def warning(text):
        error_console.print(text)

    warnings = eradiate.kernel.check_kernel()

    if warnings:
        section("Warnings", newline=False)
    for w in warnings:
        warning("• " + w)

    section("System")
    message(f"CPU: {sys_info['cpu_info']}")
    message(f"OS: {sys_info['os']}")
    message(f"Python: {sys_info['python']}")

    section("Versions")
    message(f"• eradiate {eradiate.__version__}")
    message(f"• drjit {sys_info['drjit_version']}")
    message(f"• mitsuba {sys_info['mitsuba_version']}")
    message(f"• eradiate-mitsuba {sys_info['eradiate_mitsuba_version']}")
    message(f"• numpy {sys_info['numpy']}")
    message(f"• scipy {sys_info['scipy']}")
    message(f"• xarray {sys_info['xarray']}")

    section("Available Mitsuba variants")
    message("\n".join([f"• {variant}" for variant in mi.variants()]))

    section("Configuration")
    for var in ["SOURCE_DIR", "ENV"]:
        value = getattr(eradiate.config, var)
        var_repr = str(value)
        message(f"• ERADIATE_{var.upper()}: {var_repr}")

    message("• Loaded setting files:")
    for fname in eradiate.config.settings._loaded_files:
        message(f"  • {fname}")


__doc__ = main.__doc__



from .ckd import Bin as Bin
from .ckd import BinSet as BinSet
from .ckd import QuadSpec as QuadSpec
from .ckd import QuadSpecErrorThreshold as QuadSpecErrorThreshold
from .ckd import QuadSpecFixed as QuadSpecFixed
from .ckd import QuadSpecMinError as QuadSpecMinError
from .index import CKDSpectralIndex as CKDSpectralIndex
from .index import MonoSpectralIndex as MonoSpectralIndex
from .index import SpectralIndex as SpectralIndex
from .mono import WavelengthSet as WavelengthSet



import numpy as np

from ...experiments import AtmosphereExperiment
from ...units import unit_registry as ureg


# TODO : This probably needs to go somewhere else or get called by fixtures
def absorption_database_error_handler_config():
    """
    Error handler configuration for absorption coefficient interpolation.

    Notes
    -----
    This configuration is chosen to ignore all interpolation issues (except
    bounds error along the mole fraction dimension) because warnings are
    captured by pytest which will raise.
    Ignoring the bounds on pressure and temperature is safe because
    out-of-bounds values usually correspond to locations in the atmosphere
    that are so high that the contribution to the absorption coefficient
    are negligible at these heights.
    The bounds error for the 'x' (mole fraction) coordinate is considered
    fatal.
    """
    return {
        "p": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        "t": {"missing": "raise", "scalar": "raise", "bounds": "ignore"},
        "x": {"missing": "ignore", "scalar": "ignore", "bounds": "raise"},
    }


def create_rpv_afgl1986_continental_brfpp(absorption_database_error_handler_config):
    r"""
    RPV AFGL1986 Aerosol
    ====================

    Rationale
    ---------

    This test case uses a basic atmospheric scene:

    * RPV surface emulating a canopy
    * Molecular atmosphere following the AFGL 1986 model
    * Aerosol layer at 1km elevation

    Parameters

    * Atmosphere: Molecular atmosphere using the agfl1986 profile
    * Aerosol layer: 16 layers from 1km to 2km height, :math:`\tau_{500} = 0.5`
      Radiative properties from the Govaerts 2021 Continental dataset
    * Surface: Square surface with RPV BSDF with :math:`k = 0.95`, :math:`g = -0.1`
      and :math:`\rho_0 = 0.027685`
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points)
    """

    return AtmosphereExperiment(
        surface={"type": "rpv", "k": 0.95, "g": -0.1, "rho_0": 0.027685},
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "irradiance": 20.0,
        },
        measures={
            "type": "mdistant",
            "construct": "hplane",
            "azimuth": 0.0,
            "zeniths": np.arange(-75.0, 75.01, 2.0),
            "srf": {"type": "multi_delta", "wavelengths": 550 * ureg.nm},
        },
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": {
                "type": "molecular",
                "thermoprops": {
                    "identifier": "afgl_1986-us_standard",
                    "z": np.arange(0, 120.1, 0.1) * ureg.km,
                    "additional_molecules": False,
                },
                "absorption_data": "monotropa",
                "error_handler_config": absorption_database_error_handler_config,
            },
            "particle_layers": {
                "type": "particle_layer",
                "bottom": 1 * ureg.km,
                "top": 2 * ureg.km,
                "tau_ref": 0.5,
                "dataset": "govaerts_2021-continental",
            },
        },
    )


def create_rpv_afgl1986_brfpp(absorption_database_error_handler_config):
    r"""
    RPV AFGL1986
    ============

    Rationale
    ---------

    This test case uses a basic atmospheric scene:

    * RPV surface emulating a canopy
    * Molecular atmosphere following the AFGL 1986 model

    Parameters

    * Atmosphere: Molecular atmosphere using the agfl1986 profile
    * Surface: Square surface with RPV BSDF with :math:`k = 0.95`, :math:`g = -0.1`
      and :math:`\rho_0 = 0.027685`
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points)
    """
    return AtmosphereExperiment(
        surface={"type": "rpv", "k": 0.95, "g": -0.1, "rho_0": 0.027685},
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "irradiance": 20.0,
        },
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "azimuth": 0.0,
                "zeniths": np.arange(-75.0, 75.01, 2.0),
                "srf": {"type": "multi_delta", "wavelengths": 550 * ureg.nm},
            }
        ],
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": {
                "type": "molecular",
                "thermoprops": {
                    "identifier": "afgl_1986-us_standard",
                    "z": np.linspace(0, 120, 61) * ureg.km,
                    "additional_molecules": False,
                },
                "absorption_data": "monotropa",
                "error_handler_config": absorption_database_error_handler_config,
            },
        },
    )



from __future__ import annotations

import itertools
from collections import OrderedDict

import numpy as np
import pint
import pinttrs
import xarray as xr
from pinttr.util import always_iterable

from .._mode import Mode
from ..exceptions import UnsupportedModeError
from ..kernel import bitmap_to_dataarray
from ..scenes.illumination import (
    ConstantIllumination,
    AbstractDirectionalIllumination,
    Illumination,
)
from ..scenes.spectra import InterpolatedSpectrum, Spectrum, UniformSpectrum
from ..spectral.spectral_set import SpectralSet
from ..units import symbol, to_quantity
from ..units import unit_context_config as ucc
from ..units import unit_context_kernel as uck
from ..units import unit_registry as ureg


def _spectral_dims(mode: Mode | str) -> tuple[tuple[str, dict], ...]:
    """
    Return spectral dimension names and associated metadata depending on the
    passed mode.
    """
    # Note: This is a callable because of dynamic unit symbol evaluation
    w = (
        "w",
        {
            "standard_name": "radiation_wavelength",
            "long_name": "wavelength",
            "units": symbol(ucc.get("wavelength")),
        },
    )
    g = (
        "g",
        {
            "standard_name": "g",
            "long_name": "g",
            "units": symbol(ucc.get("dimensionless")),
        },
    )
    if mode == "mono" or mode.is_mono:
        return (w,)
    elif mode == "ckd" or mode.is_ckd:
        return w, g
    else:
        raise UnsupportedModeError


def aggregate_ckd_quad(
    mode_id: str, raw_data: xr.DataArray, spectral_set: SpectralSet, is_variance: bool
) -> xr.DataArray:
    """
    Compute CKD quadrature.

    In CKD modes, this pipeline step aggregates spectral data and
    evaluates the selected quadrature rule. The following updates to the input
    data are expected:

    * the ``g`` dimension is dropped;
    * a ``bin_wmin`` (resp. ``bin_wmax``) coordinate is created and contains the
      lower (resp. upper) spectral bound of each bin;
    * the dataset is reordered by ascending ``w`` values.

    Parameters
    ----------
    mode_id : str
        Eradiate mode for which this pipeline step is configured. If set to
        "ckd", CKD quadrature computation is performed; otherwise, it is a
        no-op.

    raw_data : DataArray
        A data array holding raw bitmap data indexed against pixel indices.

    spectral_set : .WavelengthSet or .BinSet
        Spectral set for which the CKD quadrature is computed.

    is_variance : bool
        Flag that specifies whether the raw_data is a variance value.

    Returns
    -------
    DataArray
        A data array indexed against spectral coordinates associated with each
        spectral bin.

    Notes
    -----
    * In non-CKD modes, this step is a no-op.
    * During pipeline assembly, this node expands as a single node named
      ``<var>``.
    * The ``spp`` variable is averaged on the ``index`` dimension.
    """
    mode = Mode.new(mode_id)
    result_name = raw_data.name.removesuffix("_raw")

    if not mode.is_ckd:
        result = raw_data.copy()
        result.name = result_name
        return raw_data

    # Get dimensions of current variable
    sizes = OrderedDict((y, len(raw_data.coords[y])) for y in raw_data.sizes)

    for dim in ["w", "g"]:
        if dim not in sizes:
            raise ValueError(
                f"CKD quadrature computation requires dimension {dim}, missing "
                "from input data"
            )

    # Init storage
    del sizes["w"]
    del sizes["g"]

    # -- Collect wavelengths associated with each bin
    wavelength_units = ucc.get("wavelength")
    wavelengths = np.array(raw_data.coords["w"].values, dtype=float)
    bin_wmins = spectral_set.wmins.m_as(wavelength_units)
    bin_wmaxs = spectral_set.wmaxs.m_as(wavelength_units)

    # -- Proceed with actual storage initialization
    result = xr.full_like(raw_data, np.nan).isel(g=0, drop=True)

    # For each bin and each pixel, compute quadrature and store the result
    for i_bin, (bin, w) in enumerate(zip(spectral_set.bins, wavelengths)):
        values_at_nodes = raw_data.sel(w=w).values

        # Rationale: Avoid using xarray's indexing in this loop for
        # performance reasons (wrong data indexing method will result in
        # 10x+ speed reduction)
        for indexes in itertools.product(*[list(range(n)) for n in sizes.values()]):
            interval = (0.0, 1.0)

            if is_variance:
                weights = bin.quad.weights.copy()
                if interval is not None:
                    weights *= 0.5 * (interval[1] - interval[0])

                variance = values_at_nodes[(slice(None), *indexes)]
                weighted_sum = float(np.dot(weights**2, variance))
                result.values[(i_bin, *indexes)] = weighted_sum
            else:
                result.values[(i_bin, *indexes)] = bin.quad.integrate(
                    values_at_nodes[(slice(None), *indexes)],
                    interval=interval,
                )

    result.attrs = raw_data.attrs
    result.name = result_name

    # TODO: Provide separate implementation
    # # Average the 'spp' variable over the 'index' dimension
    # with xr.set_options(keep_attrs=True):
    #     result["spp"] = gathered_bitmaps.spp.mean(dim="index")

    # Add spectral coordinates
    result = result.assign_coords(
        {
            "bin_wmin": (
                "w",
                bin_wmins,
                {
                    "standard_name": "bin_wmin",
                    "long_name": "spectral bin lower bound",
                    "units": symbol(wavelength_units),
                },
            ),
            "bin_wmax": (
                "w",
                bin_wmaxs,
                {
                    "standard_name": "bin_wmax",
                    "long_name": "spectral bin upper bound",
                    "units": symbol(wavelength_units),
                },
            ),
        }
    )

    # Reorder by ascending "w"
    result = result.sortby("w")

    return result


def apply_spectral_response(spectral_data: xr.DataArray, srf: Spectrum) -> xr.DataArray:
    """
    Apply SRF weighting (a.k.a convolution) to spectral data and turn it into
    a band aggregate.

    Parameters
    ----------
    spectral_data : DataArray
        Spectral data to process.

    srf : Spectrum
        Spectral response function to apply.

    Returns
    -------
    DataArray or None
        A data array where the spectral dimension is removed after applying
        SRF weighting, or None if the SRF is a :class:`.MultiDeltaSpectrum`.
    """
    if not {"bin_wmin", "bin_wmax"}.issubset(set(spectral_data.coords.keys())):
        raise ValueError(
            "input data is missing 'bin_wmin' and/or 'bin_wmax' coordinates"
        )

    # Evaluate integral of spectral response function within selected interval
    wmin = to_quantity(spectral_data.coords["bin_wmin"]).min()
    wmax = to_quantity(spectral_data.coords["bin_wmax"]).max()
    srf_int = srf.integral(wmin, wmax)

    if isinstance(srf, InterpolatedSpectrum):
        srf_w = srf.wavelengths
    elif isinstance(srf, UniformSpectrum):
        srf_w = np.array([wmin.m_as(ureg.nm), wmax.m_as(ureg.nm)]) * ureg.nm
    else:
        raise TypeError(f"unhandled SRF type '{srf.__class__.__name__}'")

    # Evaluate integral of product of variable and SRF within selected interval
    data_w = to_quantity(spectral_data.coords["w"])

    # Spectral grid is the finest between data and SRF grids
    w_units = data_w.units
    w_m = np.array(sorted(set(data_w.m_as(w_units)) | set(srf_w.m_as(w_units))))

    # If data var has length 1 on spectral dimension, directly select
    # the value instead of using interpolation (it's a known scipy issue)
    if len(spectral_data.coords["w"]) == 1:
        # Note: The tricky thing is to recreate and extend the 'w'
        # dimension with the same axis index as in the original data
        spectral_values = spectral_data.isel(w=0, drop=True).expand_dims(
            w=w_m, axis=spectral_data.get_axis_num("w")
        )

    # Otherwise, use nearest neighbour interpolation (we assume that the
    # spectral data is constant over each spectral bin)
    else:
        spectral_values = spectral_data.interp(
            w=w_m, method="nearest", kwargs={"fill_value": "extrapolate"}
        )

    srf_values = (
        srf.eval_mono(w_m * w_units)
        .reshape([-1 if dim == "w" else 1 for dim in spectral_values.dims])
        .magnitude
    )
    assert isinstance(srf_values, np.ndarray)  # Check for leftover bugs
    var_srf_int = (spectral_values * srf_values).integrate("w")

    # Initialize storage (we want to keep all coordinate variables)
    result = xr.full_like(spectral_data, np.nan).isel(w=0, drop=True)
    result.values = var_srf_int.values / srf_int.m_as(w_units)

    # Apply SRF to variable and store result
    attrs = spectral_values.attrs.copy()
    if "standard_name" in attrs:
        attrs["standard_name"] += "_srf"
    if "long_name" in attrs:
        attrs["long_name"] += " (SRF-weighted)"

    try:
        name = spectral_data.name + "_srf"
    except TypeError as e:
        raise TypeError("expected a DataArray with a name") from e

    result.attrs = attrs
    result.name = name

    return result


def compute_albedo(
    radiosity_data: xr.DataArray, irradiance_data: xr.DataArray
) -> xr.DataArray:
    """
    Compute the bi-hemispherical reflectance (a.k.a surface albedo) from
    radiosity and irradiance records.

    Parameters
    ----------
    radiosity_data : DataArray
        Radiosity record.

    irradiance_data : DataArray
        Irradiance record.

    Returns
    -------
    dict
        A dictionary with the following content:

        ``albedo`` : :class:`xarray.DataArray`
          An albedo data array.
    """

    # We assume that all quantities are stored in kernel units
    result = radiosity_data / irradiance_data
    result.name = "albedo"
    result.attrs = {
        "standard_name": "albedo",
        "long_name": "surface albedo",
        "units": "",
    }

    return result


def compute_bidirectional_reflectance(
    radiance_data: xr.DataArray, irradiance_data: xr.DataArray
) -> dict[str, xr.DataArray]:
    """
    Compute the spectral bidirectional reflectance distribution function
    (BRDF) and bidirectional reflectance factor (BRF) from radiance and
    irradiance records.

    Parameters
    ----------
    radiance_data : DataArray
        Radiance record.

    irradiance_data : DataArray
        Irradiance record.

    Returns
    -------
    dict
        A dictionary with the following content:

        ``brdf`` : :class:`xarray.DataArray`
          A BRDF data array.
        ``brf`` : :class:`xarray.DataArray`
          A BRF data array.

    Notes
    -----
    This function is expanded to ``brdf`` and ``brf`` nodes during pipeline
    assembly.
    """
    # We assume that all quantities are stored in kernel units

    brdf_data = radiance_data / irradiance_data
    brdf_data.attrs = {
        "standard_name": "brdf",
        "long_name": "bi-directional reflection distribution function",
        "units": symbol("1/sr"),
    }
    brdf_data.name = "brdf"

    brf_data = brdf_data * np.pi
    brf_data.attrs = {
        "standard_name": "brf",
        "long_name": "bi-directional reflectance factor",
        "units": symbol("dimensionless"),
    }
    brf_data.name = "brf"

    return {"brdf": brdf_data, "brf": brf_data}


def extract_irradiance(
    mode_id: str, illumination: Illumination, spectral_set: SpectralSet
) -> dict:
    """
    Derive an irradiance dataset from the irradiance spectrum of an illuminant,
    as well as solar angles.

    Parameters
    ----------
    mode_id : str
        Eradiate mode for which this pipeline step is to be configured.

    illumination : Illumination
        The illuminant whose irradiance is to be evaluated.

    spectral_set : .WavelengthSet or .BinSet
        Spectral set driving the simulation.

    Returns
    -------
    dict
        A dictionary with the following content:

        ``irradiance`` : :class:`xarray.DataArray`
          A spectral irradiance data array compatible with processed spectral
          radiance records and suitable for deriving BRDF data.
        ``solar_angles`` : :class:`xarray.Dataset` or ``None``
          If relevant, a dataset holding solar angles.

    Notes
    -----
    This function is expanded into ``irradiance`` and ``solar_angles`` nodes
    during pipeline assembly.
    """
    mode = Mode.new(mode_id)
    k_irradiance_units = uck.get("irradiance")
    wavelengths = spectral_set.wavelengths
    wavelength_metadata = _spectral_dims("mono")[0][1]
    wavelength_units = wavelength_metadata["units"]

    def eval_illumination_spectrum(
        field_name: str, k_units: pint.Unit
    ) -> pint.Quantity:
        # Local helper function to evaluate the illumination spectrum

        spectrum: Spectrum = getattr(illumination, field_name)

        if mode.is_mono:
            return spectrum.eval_mono(wavelengths).m_as(k_units)

        elif mode.is_ckd:
            # Note: We pass here None to the g parameter because irradiance
            # spectrum data never have a g coordinate: the g parameter is simply
            # ignored.
            result = spectrum.eval_ckd(w=wavelengths, g=None).m_as(k_units)

            # Reorder data by ascending wavelengths
            indices = wavelengths.argsort()

            return result[indices]

        else:
            raise UnsupportedModeError(supported=("monochromatic", "ckd"))

    # Now, proceed with actual spectrum evaluation
    solar_angles = None

    # Define spectral coordinates
    spectral_coords = {
        "w": ("w", wavelengths.m_as(wavelength_units), wavelength_metadata)
    }

    if mode.is_ckd:
        bin_wmins = spectral_set.wmins.m_as(wavelength_units)
        bin_wmaxs = spectral_set.wmaxs.m_as(wavelength_units)
        spectral_coords.update(
            {
                "bin_wmin": (
                    "w",
                    bin_wmins,
                    {
                        "standard_name": "bin_wmin",
                        "long_name": "spectral bin lower bound",
                        "units": symbol(wavelength_units),
                    },
                ),
                "bin_wmax": (
                    "w",
                    bin_wmaxs,
                    {
                        "standard_name": "bin_wmax",
                        "long_name": "spectral bin upper bound",
                        "units": symbol(wavelength_units),
                    },
                ),
            }
        )

    if isinstance(illumination, AbstractDirectionalIllumination):
        # Collect illumination angular data
        saa = illumination.azimuth.m_as(ureg.deg)
        sza = illumination.zenith.m_as(ureg.deg)
        cos_sza = np.cos(np.deg2rad(sza))

        # Package solar angles into a dataset
        solar_angles = xr.Dataset(
            {
                "sza": (
                    "sza",
                    [sza],
                    {
                        "standard_name": "solar_zenith_angle",
                        "long_name": "solar zenith angle",
                        "units": symbol("deg"),
                    },
                ),
                "saa": (
                    "saa",
                    [saa],
                    {
                        "standard_name": "solar_azimuth_angle",
                        "long_name": "solar azimuth angle",
                        "units": symbol("deg"),
                    },
                ),
            }
        )

        # Collect illumination spectral data
        irradiances = eval_illumination_spectrum("irradiance", k_irradiance_units)

        # Create irradiance variable
        irradiance = xr.DataArray(
            data=(irradiances * cos_sza).reshape((1, 1, len(irradiances))),
            dims=("sza", "saa", "w"),
            coords={
                "sza": solar_angles["sza"],
                "saa": solar_angles["saa"],
                **spectral_coords,
            },
            name="irradiance",
        )

    elif isinstance(illumination, ConstantIllumination):
        # Collect illumination spectral data
        k_radiance_units = uck.get("radiance")
        radiances = eval_illumination_spectrum("radiance", k_radiance_units)

        # Create irradiance variable
        irradiance = xr.DataArray(
            data=np.pi * radiances.reshape((len(radiances),)),
            dims=("w",),
            coords=spectral_coords,
            name="irradiance",
        )

    else:
        raise TypeError(
            "field 'illumination' must be one of "
            "(DirectionalIllumination, ConstantIllumination), got a "
            f"{illumination.__class__.__name__}"
        )

    irradiance.attrs = {
        "standard_name": "horizontal_solar_irradiance_per_unit_wavelength",
        "long_name": "horizontal spectral irradiance",
        "units": symbol(k_irradiance_units),
    }

    return {"irradiance": irradiance, "solar_angles": solar_angles}


def gather_bitmaps(
    mode_id: str,
    var_name: str,
    var_metadata: dict,
    gather_variance: bool,
    bitmaps: dict,
    viewing_angles: xr.Dataset,
    solar_angles: xr.Dataset,
) -> dict:
    """
    Gather a collection of Mitsuba bitmaps into xarray data arrays.

    Parameters
    ----------
    mode_id : str
        Eradiate mode from which this pipeline step is configured.

    var_name : str
        Name of the processed physical variable.

    var_metadata : dict
        A metadata dictionary to be attached to the data array holding the
        processed physical variable.

    gather_variance : bool
        Flag that specifies whether the variance bitmaps should be gathered.

    bitmaps : dict
        A dictionary mapping spectral loop indexes to the corresponding bitmap.
        Dictionary structure is as follows:

        .. code::

           {
               <loop_index_0>: {
                   {"bitmap": <bitmap_0>},
                   {"spp": <sample_count_0>},
               },
               <loop_index_1>: {
                   {"bitmap": <bitmap_1>},
                   {"spp": <sample_count_1>},
               },
               ...
           }

    viewing_angles : Dataset, optional
        A dataset holding the viewing angles associated with each pixel in the
        processed bitmaps.

    solar_angles : Dataset, optional
        A dataset holding the solar angles associated with the processed
        observation data.

    Returns
    -------
    data_vars : dict[str, DataArray]
        A dictionary mapping data variable names to a corresponding data array.
        These can easily be aggregated into an xarray dataset, or scattered
        around other nodes of the post-processing pipeline.
    """
    mode = Mode.new(mode_id)

    # Set up spectral dimensions
    spectral_dims = []
    spectral_dim_metadata = {}

    for y in _spectral_dims(mode):
        if isinstance(y, str):
            spectral_dims.append(y)
            spectral_dim_metadata[y] = {}
        else:
            spectral_dims.append(y[0])
            spectral_dim_metadata[y[0]] = y[1]

    # Loop on spectral indexes and collect all bitmap contents in data arrays
    sensor_data = {
        "spp": [],
        "weights_raw": [],
        f"{var_name}_raw": [],
        f"{var_name}_m2_raw": [],
    }

    for spectral_index_hashable, result_dict in bitmaps.items():
        spectral_index = spectral_index_hashable

        da = bitmap_to_dataarray(result_dict["bitmap"])
        spp = result_dict["spp"]

        # Set spectral coordinates
        all_coords = {
            spectral_dim: [spectral_coord]
            for spectral_dim, spectral_coord in zip(
                spectral_dims, always_iterable(spectral_index)
            )
        }

        # Add spectral and sensor dimensions to img array
        sensor_data[f"{var_name}_raw"].append(da.expand_dims(dim=all_coords))

        # Package spp in a data array
        all_dims = list(all_coords.keys())
        spp_shape = [1 for _ in all_dims]
        sensor_data["spp"].append(
            xr.DataArray(np.reshape(spp, spp_shape), coords=all_coords)
        )

        if gather_variance:
            da_var = bitmap_to_dataarray(result_dict["m2"])
            sensor_data[f"{var_name}_m2_raw"].append(da_var.expand_dims(dim=all_coords))

    # Combine all the data
    result = {k: xr.combine_by_coords(v) if v else None for k, v in sensor_data.items()}

    keys = [f"{var_name}_raw"]
    if gather_variance:
        keys.append(f"{var_name}_m2_raw")

    for key in keys:
        # Add viewing angle coordinates to main data array
        if viewing_angles is not None:
            result[key] = result[key].assign_coords(viewing_angles)

        # Add solar angle coordinate to main data array
        if solar_angles is not None:
            result[key] = result[key].expand_dims(
                {k: solar_angles[k].values for k in ["sza", "saa"]}, axis=(-1, -2)
            )

    # Drop "channel" dimension when using a monochromatic Mitsuba variant
    if mode.check(mi_color_mode="mono"):
        for k, v in result.items():
            if v is not None and "channel" in v.dims:
                result[k] = v.squeeze("channel", drop=True)

    # Apply spectral metadata
    for da in result.values():
        if da is not None:
            for spectral_dim in spectral_dims:
                da[spectral_dim].attrs = spectral_dim_metadata[spectral_dim]

    # Update name of all variables
    for name, da in result.items():
        if da is not None:
            da.name = name

    # Apply metadata to data variables
    result[f"{var_name}_raw"].attrs.update(var_metadata)

    return result


def radiosity(sector_radiosity: xr.DataArray) -> xr.DataArray:
    """
    Aggregate sector radiosity into a full-hemisphere radiosity dataset.

    Parameters
    ----------
    sector_radiosity : DataArray
        Radiosity values for the hemisphere sector corresponding to each film
        pixel. This quantity is expressed in spectral flux units
        (typically W/m²/nm) and, when summed over the entire film, aggregates
        into a spectral radiosity (W/m²/nm).

    Returns
    -------
    DataArray
        A global radiosity record, with no film pixel / viewing angle indexing
        left.
    """
    result = sector_radiosity.sum(dim=("x_index", "y_index"))
    result.attrs = {
        "standard_name": "toa_outgoing_flux_density_per_unit_wavelength",
        "long_name": "top-of-atmosphere outgoing spectral flux density",
        "units": symbol(uck.get("irradiance")),
    }

    return result


def spectral_response(srf: Spectrum) -> xr.DataArray:
    """
    Evaluate a spectral response function as a data array.

    Parameters
    ----------
    srf : Spectrum
        The spectral function to be evaluated.

    Returns
    -------
    DataArray
        A data array mapping SRF values against the wavelength. To avoid
        confusion with the spectral coordinate, the wavelength dimension is here
        named ``srf_w``.
    """
    # Evaluate SRF
    w_units = ucc.get("wavelength")

    if isinstance(srf, InterpolatedSpectrum):
        srf_w = srf.wavelengths
        srf_values = pinttrs.util.ensure_units(srf.values, ureg.dimensionless)

    else:
        raise TypeError(f"unsupported SRF type '{srf.__class__.__name__}'")

    result = xr.DataArray(
        data=srf_values.m,
        dims=("srf_w",),
        coords={
            "srf_w": (
                "srf_w",
                srf_w.m_as(w_units),
                {
                    "standard_name": "radiation_wavelength",
                    "long_name": "wavelength",
                    "units": symbol(w_units),
                },
            )
        },
        attrs={
            "standard_name": "spectral_response_function",
            "long_name": "spectral response function",
            "units": "",
        },
    )

    return result


def viewing_angles(angles: np.ndarray) -> xr.Dataset:
    """
    Collect viewing angles associated with each film pixel from the measure, if
    relevant.

    Parameters
    ----------
    angles : ndarray
        An array with shape (m, n, 2) containing the processed measure's viewing
        angles in degree. Typically obtained by calling
        :meth:`.DistantMeasure.viewing_angles`.

    Returns
    -------
    Dataset
        An xarray dataset holding viewing angle values indexed by film
        coordinates.
    """
    theta = angles[:, :, 0]
    phi = angles[:, :, 1]

    return xr.Dataset(
        {
            "vza": xr.DataArray(
                theta,
                coords={
                    "x_index": [i for i in range(theta.shape[0])],
                    "y_index": [i for i in range(theta.shape[1])],
                },
                dims=("x_index", "y_index"),
                attrs={
                    "standard_name": "viewing_zenith_angle",
                    "long_name": "viewing zenith angle",
                    "units": symbol("deg"),
                },
            ),
            "vaa": xr.DataArray(
                phi,
                coords={
                    "x_index": [i for i in range(phi.shape[0])],
                    "y_index": [i for i in range(phi.shape[1])],
                },
                dims=("x_index", "y_index"),
                attrs={
                    "standard_name": "viewing_azimuth_angle",
                    "long_name": "viewing azimuth angle",
                    "units": symbol("deg"),
                },
            ),
        }
    )


def moment2_to_variance(
    expectation: xr.DataArray, m2: xr.DataArray, spp: xr.DataArray
) -> xr.DataArray:
    """
    Calculate the variance (central 2nd moment) from the raw 2nd moment and expected value.

    Parameters
    ----------
    expectation : DataArray
        The data expected value. For Monte Carlo integration, this is the output of
        the estimator.

    m2 : DataArray
        The data's raw 2nd moment.

    spp : DataArray
        The number of samples per pixels used during the Monte Carlo integration.

    Returns
    -------
    DataArray
        The variance (central 2nd moment) of the data. The DataArray has the same shape
        as the expectation or m2.

    Notes
    -----
    This function calculates the Monte-Carlo variance, which aggregates the variance
    of each sample. It thus needs to be divided by the spp.
    """
    variance = (m2 - expectation * expectation) / spp
    variance.name = m2.name.replace("m2", "var")
    return variance



from __future__ import annotations

import attrs
import mitsuba as mi
import pint
import pinttr

from ._core import BSDF
from ..core import traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import documented, parse_docs
from ...kernel import TypeIdLookupStrategy, UpdateParameter
from ...units import unit_context_config as ucc


@parse_docs
@attrs.define(eq=False, slots=False)
class HapkeBSDF(BSDF):
    """
    Hapke BSDF [``hapke``].

    This BSDF implements the Hapke surface model as described in
    :cite:`Hapke1984BidirectionalReflectanceSpectroscopy`. This highly flexible
    and robust surface model allows for the characterisation of a sharp
    back-scattering hot spot. The so-called Hapke model has been adapted to
    several different use cases in the litterature, the version with 6
    parameters implemented here is one of the most commonly used.

    See Also
    --------
    :ref:`plugin-bsdf-hapke`
    """

    w: Spectrum = documented(
        attrs.field(
            default=None,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Single scattering albedo 'w'. Must be in [0; 1]",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
    )

    b: Spectrum = documented(
        attrs.field(
            default=None,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Anisotropy parameter 'b' Must be in [0; 1]",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
    )

    c: Spectrum | None = documented(
        attrs.field(
            default=None,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Scattering coefficient 'c'. Must be in [0; 1]",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
    )

    theta: Spectrum = documented(
        attrs.field(
            default=0.183,
            converter=spectrum_factory.converter("angle"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("angle"),
            ],
        ),
        doc="Photometric roughness 'theta'. Angle in degree. Must be in [0; 90]°",
        type=".Spectrum",
        init_type="quantity or float",
    )

    B_0: Spectrum = documented(
        attrs.field(
            default=None,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Shadow hiding opposition effect amplitude 'B_0'. Must be in [0; 1]",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
    )

    h: Spectrum = documented(
        attrs.field(
            default=None,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="shadow hiding opposition effect width 'h'. Must be in [0; 1]",
        type=".Spectrum",
        init_type=".Spectrum or dict or float",
    )

    @property
    def template(self) -> dict:
        # Inherit docstring
        objects = {
            "w": traverse(self.w)[0],
            "b": traverse(self.b)[0],
            "c": traverse(self.c)[0],
            "theta": traverse(self.theta)[0],
            "B_0": traverse(self.B_0)[0],
            "h": traverse(self.h)[0],
        }

        result = {
            "type": "hapke"
        }

        for obj_key, obj_values in objects.items():
            for key, value in obj_values.items():
                result[f"{obj_key}.{key}"] = value

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, UpdateParameter]:
        # Inherit docstring
        objects = {
            "w": traverse(self.w)[1],
            "b": traverse(self.b)[1],
            "c": traverse(self.c)[1],
            "theta": traverse(self.theta)[1],
            "B_0": traverse(self.B_0)[1],
            "h": traverse(self.h)[1],
        }

        result = {}
        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{obj_key}.{key}"] = attrs.evolve(
                    param,
                    lookup_strategy=TypeIdLookupStrategy(
                        node_type=mi.BSDF,
                        node_id=self.id,
                        parameter_relpath=f"{obj_key}.{key}",
                    )
                    if self.id is not None
                    else None,
                )

        return result



from __future__ import annotations

import logging
import os
import shutil
import typing as t
from pathlib import Path, PurePosixPath

import attrs
import pooch

from ._core import DataStore, expand_rules, registry_from_file
from ..exceptions import DataError
from ..typing import PathLike
from ..util.misc import LoggingContext

logger = logging.getLogger(__name__)


@attrs.define(repr=False, init=False)
class SafeOnlineDataStore(DataStore):
    """
    Serve files located online, with integrity check.

    Parameters
    ----------
    base_url : str
        URL to the online storage location.

    path : path-like
        Path to the local cache location.

    registry_fname : path-like, optional
        Path to the registry file, relative to `path`.

    attempts : int, default: 3
        Number of download attempts to make before giving up because of
        connection errors or a hash mismatch.

    Fields
    ------
    manager : pooch.Pooch
        The Pooch instance used to manage downloaded content.

    registry_fname : Path
        Path to the registry file, relative to `path`.

    Notes
    -----
    This class basically wraps a :class:`pooch.Pooch` instance.
    """

    manager: pooch.Pooch = attrs.field()
    registry_fname: Path = attrs.field(converter=Path)

    def __init__(
        self,
        base_url: str,
        path: PathLike,
        registry_fname: PathLike = "registry.txt",
        attempts: int = 3,
    ):
        # Initialize attributes
        if not base_url.endswith("/"):
            base_url += "/"
        path = Path(path).absolute()

        manager = pooch.create(
            base_url=base_url,
            path=path,
            registry=None,  # We'll load it later
            retry_if_failed=attempts - 1,
        )
        self.__attrs_init__(manager=manager, registry_fname=registry_fname)

        # Initialize register load the registry
        registry = registry_from_file(self.registry_fetch())
        manager.registry = registry

    def __repr__(self):
        attr_reprs = [
            f"{x}={self.__getattribute__(x).__repr__()}"
            for x in [
                "base_url",
                "path",
            ]
        ]

        return f"SafeOnlineDataStore({', '.join(attr_reprs)})"

    @property
    def base_url(self) -> str:
        # Inherit docstring
        return self.manager.base_url

    @property
    def path(self) -> Path:
        """
        path : Absolute path to the local data storage folder.
        """
        return Path(self.manager.path)

    @property
    def registry(self) -> dict[str, str]:
        # Inherit docstring
        return self.manager.registry

    @property
    def retry_if_failed(self):
        return self.manager.retry_if_failed

    def registry_files(
        self, filter: t.Callable[[t.Any], bool] | None = None
    ) -> list[str]:
        """
        Get a list of registered files.

        Parameters
        ----------
        filter : callable, optional
            A filter function taking a file path as a single string argument and
            returning a Boolean. Filenames for which the filter returns ``True``
            will be returned.

        Returns
        -------
        files : list of str
            List of registered files.
        """
        if filter is None:
            return self.manager.registry_files
        else:
            return [x for x in self.manager.registry_files if filter(x)]

    @property
    def registry_path(self) -> Path:
        """
        Path: Absolute path to the registry file.
        """
        return self.path / self.registry_fname

    def registry_fetch(self) -> Path:
        """
        Get the absolute path to the registry file and make sure that it is
        written to the local cache.
        """
        filename = self.registry_path

        with LoggingContext(
            pooch.get_logger(), level="WARNING"
        ):  # Silence pooch messages temporarily
            result = pooch.retrieve(
                os.path.join(self.base_url, self.registry_fname),
                known_hash=None,
                fname=str(filename),
                path=self.path,
            )

        return Path(result)

    def registry_delete(self):
        """
        Delete the registry file.
        """
        os.remove(self.path / self.registry_fname)

    def registry_reload(self, delete: bool = False) -> None:
        """
        Reload the registry file from the local cache.

        Parameters
        ----------
        delete : bool, optional
            If ``True``, the existing registry file will be deleted and
            downloaded again.
        """
        if delete:
            self.registry_delete()

        registry_fname = self.registry_fetch()
        self.manager.registry = registry_from_file(registry_fname)

    def is_registered(
        self, filename: PathLike, allow_compressed: bool = True
    ) -> PurePosixPath:
        """
        Check if a file is registered, with an option to look for compressed
        data.

        Parameters
        ----------
        filename : path-like
            File name to fetch from the local storage, relative to the storage
            root.

        allow_compressed : bool, optional
            If ``True``, a query for ``foo.bar`` will result in a query for the
            gzip-compressed file name ``foo.bar.gz``. The compressed file takes
            precedence.

        Returns
        -------
        path : Path
            The file name which matched `filename`.

        Raises
        ------
        ValueError
            If `filename` could not be matched with any entry in the registry.
        """
        fname = Path(filename).as_posix()

        if allow_compressed and not fname.endswith(".gz"):
            fname_compressed = fname + ".gz"

            if fname_compressed in self.manager.registry:
                return PurePosixPath(fname_compressed)

        if fname in self.manager.registry:
            return PurePosixPath(fname)

        raise ValueError(f"File '{fname}' is not in the registry.")

    def fetch(self, filename: PathLike, downloader: t.Callable | None = None) -> Path:
        """
        Fetch a file from the data store. This method wraps
        :meth:`pooch.Pooch.fetch` and automatically selects compressed files
        when they are available.

        Parameters
        ----------
        filename : path-like
            File name to fetch from the local storage, relative to the storage
            root.

        downloader : callable, optional
            A callable that will be called to download a given URL to a provided
            local file name. This is mostly useful to
            `display progress bars <https://www.fatiando.org/pooch/latest/progressbars.html>`_
            during download.

        Returns
        -------
        path : Path
            Absolute path where the retrieved resource is located.

        Notes
        -----
        If a compressed resource exists, it will be served automatically.
        For instance, if ``"foo.nc"`` is requested and ``foo.nc.gz`` is
        registered, the latter will be downloaded, decompressed and served as
        ``"foo.nc"``.
        """
        # By default, just forward arguments
        fname = Path(filename).as_posix()
        processor = None

        # Look up the file in the registry
        # (also detects if a gzip-compressed resource is available)
        try:
            fname = str(self.is_registered(filename))

        except ValueError as e:
            raise DataError(
                f"file '{fname}' could not be retrieved from {self.base_url}"
            ) from e

        # If the matched registered resource is a compressed file, serve it
        root, ext = os.path.splitext(fname)
        if ext == ".gz":
            processor = pooch.processors.Decompress(name=os.path.basename(root))

        return Path(
            self.manager.fetch(fname, processor=processor, downloader=downloader)
        )

    def purge(self, keep: None | str | list[str] = None) -> None:
        """
        Purge local storage location. The default behaviour is very aggressive
        and will wipe out the entire directory contents.

        Parameters
        ----------
        keep : "registered" or list of str, optional
            If set to ``"registered"``, files in the registry, as well as the
            registry file itself, will not be deleted. Finer control is possible
            by passing a list of exclusion rules (paths relative to the store's
            local storage root, shell wildcards allowed).

        Notes
        -----
        Passing ``keep="registered"`` keeps registered files to minimize the
        amount of data to be downloaded upon future queries the the data store.
        This means, for instance, that if data is registered and downloaded as
        a compressed file, then served decompressed, the compressed file will be
        kept, while the decompressed file will be deleted.

        Warnings
        --------
        This is a destructive operation, make sure you know what you're doing!
        """
        # Fast track for simple case
        if keep is None:
            for filename in os.scandir(self.path):
                file_path = os.path.join(self.path, filename)

                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.remove(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)

            return

        # List files to keep and delete
        if keep == "registered":
            excluded = expand_rules(
                rules=self.registry_files() + [str(self.registry_fname)],
                prefix=self.path,
            )
        else:
            excluded = expand_rules(rules=keep, prefix=self.path)
        included = expand_rules(rules=["**/*"], prefix=self.path)
        remove = sorted(included - excluded)

        for file in remove:
            os.remove(file)

        # Clean up empty directories
        for x in self.path.iterdir():
            if x.is_dir():
                try:
                    os.removedirs(x)
                except OSError:
                    pass



from __future__ import annotations

import os
from abc import ABC, abstractmethod

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr

from ..core import CompositeSceneElement
from ... import validators
from ..._factory import Factory
from ...attrs import documented, get_doc, parse_docs
from ...kernel import UpdateParameter
from ...typing import PathLike
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ...util.misc import flatten

biosphere_factory = Factory()
biosphere_factory.register_lazy_batch(
    [
        (
            "_core.InstancedCanopyElement",
            "instanced",
            {},
        ),
        (
            "_discrete.DiscreteCanopy",
            "discrete_canopy",
            {"dict_constructor": "padded"},
        ),
        (
            "_leaf_cloud.LeafCloud",
            "leaf_cloud",
            {},
        ),
        (
            "_tree.AbstractTree",
            "abstract_tree",
            {},
        ),
        (
            "_tree.MeshTree",
            "mesh_tree",
            {},
        ),
    ],
    cls_prefix="eradiate.scenes.biosphere",
)


@parse_docs
@attrs.define(eq=False, slots=False)
class Canopy(CompositeSceneElement, ABC):
    """
    Abstract base class for all canopies.
    """

    id: str | None = documented(
        attrs.field(
            default="canopy",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(CompositeSceneElement, "id", "doc"),
        type=get_doc(CompositeSceneElement, "id", "type"),
        init_type=get_doc(CompositeSceneElement, "id", "init_type"),
        default='"canopy"',
    )

    size: pint.Quantity | None = documented(
        pinttr.field(
            default=None,
            validator=attrs.validators.optional(
                [
                    pinttr.validators.has_compatible_units,
                    validators.on_quantity(validators.is_vector3),
                ]
            ),
            units=ucc.deferred("length"),
        ),
        doc="Canopy extent as a 3-vector.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
    )


@parse_docs
@attrs.define(eq=False, slots=False)
class CanopyElement(CompositeSceneElement, ABC):
    """
    Abstract base class representing a component of a :class:`.Canopy` object.
    Concrete canopy classes can manage their components as they prefer.
    """

    @property
    @abstractmethod
    def _template_bsdfs(self) -> dict:
        pass

    @property
    @abstractmethod
    def _template_shapes(self) -> dict:
        pass

    @property
    def template(self) -> dict:
        return flatten({**self._template_bsdfs, **self._template_shapes})

    @property
    @abstractmethod
    def _params_bsdfs(self) -> dict:
        pass

    @property
    @abstractmethod
    def _params_shapes(self) -> dict:
        pass

    @property
    def params(self) -> dict[str, UpdateParameter]:
        # Inherit docstring
        return flatten({**self._params_bsdfs, **self._params_shapes})


@parse_docs
@attrs.define(eq=False, slots=False)
class InstancedCanopyElement(CompositeSceneElement):
    """
    Instanced canopy element [``instanced``].

    This class wraps a canopy element and defines locations where to position
    instances (*i.e.* clones) of it.

    .. admonition:: Class method constructors

       .. autosummary::

          from_file
    """

    canopy_element: CanopyElement | None = documented(
        attrs.field(
            default=None,
            validator=attrs.validators.optional(
                attrs.validators.instance_of(CanopyElement)
            ),
            converter=biosphere_factory.convert,
        ),
        doc="Instanced canopy element. Can be specified as a dictionary, which "
        "will be converted by :data:`.biosphere_factory`.",
        type=":class:`.CanopyElement`, optional",
    )

    instance_positions: pint.Quantity = documented(
        pinttr.field(
            factory=list,
            units=ucc.deferred("length"),
        ),
        doc="Instance positions as an (n, 3)-array.\n"
        "\n"
        "Unit-enabled field (default: ucc['length'])",
        type="quantity",
        init_type="array-like",
        default="[]",
    )

    @instance_positions.validator
    def _instance_positions_validator(self, attribute, value):
        if value.shape and value.shape[0] > 0 and value.shape[1] != 3:
            raise ValueError(
                f"while validating {attribute.name}, must be an array of shape "
                f"(n, 3), got {value.shape}"
            )

    # --------------------------------------------------------------------------
    #                               Constructors
    # --------------------------------------------------------------------------

    @classmethod
    def from_file(
        cls,
        filename: PathLike,
        canopy_element: CanopyElement | None = None,
    ):
        """
        Construct a :class:`.InstancedCanopyElement` from a text file specifying
        instance positions.

        .. admonition:: File format

           Each line defines an instance position as a whitespace-separated
           3-vector of Cartesian coordinates.

        .. important::

           Location coordinates are assumed to be given in meters.

        Parameters
        ----------
        filename : path-like
            Path to the text file specifying the leaves in the canopy.
            Can be absolute or relative.

        canopy_element : .CanopyElement or dict, optional
            :class:`.CanopyElement` to be instanced. If a dictionary is passed,
            if is interpreted by :data:`.biosphere_factory`. If set to
            ``None``, an empty leaf cloud will be created.

        Returns
        -------
        :class:`.InstancedCanopyElement`
            Created :class:`.InstancedCanopyElement`.

        Raises
        ------
        ValueError
            If ``filename`` is set to ``None``.

        FileNotFoundError
            If ``filename`` does not point to an existing file.
        """
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"no file at {filename} found.")

        if canopy_element is None:
            canopy_element = {"type": "leaf_cloud"}

        canopy_element = biosphere_factory.convert(canopy_element)

        instance_positions = []

        with open(filename, "r") as f:
            for i_line, line in enumerate(f):
                try:
                    coords = np.array(line.split(), dtype=float)
                except ValueError as e:
                    raise ValueError(
                        f"while reading {filename}, on line {i_line + 1}: "
                        f"cannot convert {line} to a 3-vector!"
                    ) from e

                if len(coords) != 3:
                    raise ValueError(
                        f"while reading {filename}, on line {i_line + 1}: "
                        f"cannot convert {line} to a 3-vector!"
                    )

                instance_positions.append(coords)

        instance_positions = np.array(instance_positions) * ureg.m
        return cls(canopy_element=canopy_element, instance_positions=instance_positions)

    # --------------------------------------------------------------------------
    #                        Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def _template_bsdfs(self) -> dict:
        # Produces the part of the kernel dictionary template describing
        # the BSDFs of the encapsulated canopy element
        return self.canopy_element._template_bsdfs

    @property
    def _template_shapes(self) -> dict:
        # Produces the part of the kernel dictionary template describing
        # the shape group of the encapsulated canopy element
        result = {f"{self.canopy_element.id}.type": "shapegroup"}

        for key, param in self.canopy_element._template_shapes.items():
            result[f"{self.canopy_element.id}.{key}"] = param

        return result

    @property
    def _template_instances(self) -> dict:
        # Produces the part of the kernel dictionary template describing
        # instances of the encapsulated canopy element
        length_units = uck.get("length")

        result = {}

        for i, position in enumerate(self.instance_positions.m_as(length_units)):
            result[f"{self.canopy_element.id}_instance_{i}.type"] = "instance"
            result[f"{self.canopy_element.id}_instance_{i}.group.type"] = "ref"
            result[
                f"{self.canopy_element.id}_instance_{i}.group.id"
            ] = self.canopy_element.id
            result[
                f"{self.canopy_element.id}_instance_{i}.to_world"
            ] = mi.ScalarTransform4f.translate(position)

        return result

    @property
    def template(self) -> dict:
        # Inherit docstring
        return {
            **self._template_bsdfs,
            **self._template_shapes,
            **self._template_instances,
        }

    @property
    def _params_bsdfs(self) -> dict:
        # Produces the part of the parameter map describing the BSDFs of the
        # encapsulated canopy element
        return self.canopy_element._params_bsdfs

    @property
    def _params_shapes(self) -> dict:
        # Produces the part of the parameter map describing the shape group of
        # the encapsulated canopy element
        return self.canopy_element._params_shapes

    @property
    def _params_instances(self) -> dict:
        # Produces the part of the parameter map describing instances of the
        # encapsulated canopy element
        return {}

    @property
    def params(self) -> dict:
        # Inherit docstring
        return {
            **self._params_bsdfs,
            **self._params_shapes,
            **self._params_instances,
        }



from __future__ import annotations

import typing as t
from collections.abc import MutableMapping

import attrs
import dessinemoi
import pinttr

from .units import unit_registry as ureg


@attrs.define
class Factory(dessinemoi.Factory):
    """
    Object factory.

    This factory subclass extends the original :class:`dessinemoi.Factory` type
    by adding the following functionality to its :meth:`convert` method:

    * unit interpretation using :func:`pinttr.interpret_units`;
    * class method constructor selection using the ``"construct"`` key
      (overrides the default dict constructor if any; reserves the
      ``"construct"`` parameter).

    The interface is unchanged.
    """

    def _convert_impl(
        self,
        value,
        allowed_cls: type | tuple[type] | None = None,
    ) -> t.Any:
        if isinstance(value, MutableMapping):
            # Interpret units and copy value to avoid unintended mutation
            value_copy = pinttr.interpret_units(value, ureg=ureg)

            # Query registry
            type_id = value_copy.pop("type")

            try:
                entry = self.registry[type_id]
            except KeyError as e:
                raise ValueError(f"no type registered as '{type_id}'") from e

            # Resolve lazy type if necessary
            cls = (
                entry.cls.load()
                if isinstance(entry.cls, dessinemoi.LazyType)
                else entry.cls
            )

            # Check if class is allowed
            if allowed_cls is not None and not issubclass(cls, allowed_cls):
                raise TypeError(
                    f"conversion to object type '{type_id}' ({cls}) is not allowed"
                )

            # Get constructor from dict, if any
            dict_constructor = value_copy.pop("construct", entry.dict_constructor)

            # Construct object
            return self.create(type_id, construct=dict_constructor, kwargs=value_copy)

        else:
            # Check if object has allowed type
            if allowed_cls is not None:
                if not isinstance(value, allowed_cls):
                    raise TypeError("value type is not allowed")

            return value

    def register_lazy_batch(
        self,
        specs: list[tuple[str, str, dict]],
        cls_prefix: str = "",
    ):
        """
        Register multiple lazy types at once.

        Parameters
        ----------
        specs : list of tuple[str, str, dict]
            A list of ``(cls, type_id, kwargs)`` tuples where ``cls`` is  the
            name of the target type (relative to `cls_prefix`), ``type_id`` is
            the ID for the registered type and ``kwargs`` is a dictionary
            containing keyword arguments for the :meth:`.Factory.register`
            method.

        cls_prefix : str
            A prefix relative to which lazy type names are expressed.
        """
        for cls, type_id, kwargs in specs:
            if cls_prefix:
                cls = f"{cls_prefix}.{cls}"
            self.register(cls, type_id=type_id, **kwargs)



from __future__ import annotations

from collections import OrderedDict

from ._blind_directory import BlindDirectoryDataStore
from ._blind_online import BlindOnlineDataStore
from ._multi import MultiDataStore
from ._safe_directory import SafeDirectoryDataStore
from ._safe_online import SafeOnlineDataStore
from .. import config

#: Global data store.
data_store: MultiDataStore | None = None


def init_data_store(
    offline: bool | None = None, production: bool | None = None
) -> None:
    """
    Initialize the global data store.

    Parameters
    ----------
    offline : bool, optional
        If ``True``, replace all online data stores with blind directory data
        stores. If unset, the global offline configuration is used.

    production : bool, optional
        If ``True``, replace all development data stores providing files from
        the current local source folder with online safe data stores. If set to
        ``True`` with ``offline`` flag set to ``True``, then the data must be
        provided by the user before accessing it. If unset, the global
        production configuration is used.

    Notes
    -----
    This function is called automatically when the ``eradiate.data`` package is
    imported.
    """
    global data_store

    download_dir = config.settings.download_dir
    source_dir = config.SOURCE_DIR
    data_dir = None if source_dir is None else config.SOURCE_DIR / "resources" / "data"

    if offline is None:
        offline = config.settings.offline

    if production is None:
        production = config.SOURCE_DIR is None

    # 'small_files' configuration
    if offline:
        if production:
            # Offline, prod: we expect files to be in the download directory
            small_files = BlindDirectoryDataStore(path=download_dir)
        else:
            # Offline, dev: we expect files to be in the data submodule
            small_files = SafeDirectoryDataStore(path=data_dir)
    else:
        if production:
            # Online, prod: we expect files to be on GitHub
            small_files = SafeOnlineDataStore(
                base_url="/".join(
                    [
                        config.settings.small_files_registry_url,
                        config.settings.small_files_registry_revision,
                    ]
                ),
                path=download_dir,
            )
        else:
            # Online, dev: we expect files to be in the data submodule
            small_files = SafeDirectoryDataStore(path=data_dir)

    # 'large_files_stable' configuration
    if offline:
        # Offline: we expect files to be in the download directory
        large_files_stable = BlindDirectoryDataStore(path=download_dir / "stable")
    else:
        # Online: we expect files to be on the Internet, with checksum
        large_files_stable = SafeOnlineDataStore(
            base_url="/".join([config.settings.data_store_url, "stable"]),
            path=download_dir / "stable",
        )

    # 'large_files_unstable' configuration
    if offline:
        # Offline: we expect files to be in the download directory
        large_files_unstable = BlindDirectoryDataStore(path=download_dir / "unstable")
    else:
        # Online: we expect files to be on the Internet, without checksum
        large_files_unstable = BlindOnlineDataStore(
            base_url="/".join([config.settings.data_store_url, "unstable"]),
            path=download_dir / "unstable",
        )

    data_store = MultiDataStore(
        stores=OrderedDict(
            [
                ("small_files", small_files),
                ("large_files_stable", large_files_stable),
                ("large_files_unstable", large_files_unstable),
            ]
        )
    )


# Initialise the data store upon module import
init_data_store()



"""
Notebook support components.
"""


def install(*extensions):
    """
    Install notebook helpers.

    Parameters
    ----------
    *extensions : str
        Extensions to activate. Available extensions (all are active if this
        parameter is unset):

        * :monobold:`kernel_logging`: Route kernel logs through standard logging
          facilities. Progress display uses `tqdm <https://tqdm.github.io/>`_.
        * :monobold:`rich_pretty`: Install :func:`rich.pretty <rich.pretty.install>`
          with default configuration to the current Python REPL.
        * :monobold:`rich_traceback`:
          Install :func:`rich.traceback <rich.traceback.install>`
          with default configuration to the current Python REPL.
    """
    from rich.pretty import install as install_rich_pretty
    from rich.traceback import install as install_rich_traceback

    from eradiate.kernel import install_logging as install_kernel_logging

    if not extensions:
        extensions = {"kernel_logging", "rich_pretty", "rich_traceback"}

    if "kernel_logging" in extensions:
        install_kernel_logging()

    if "rich_pretty" in extensions:
        install_rich_pretty()

    if "rich_traceback" in extensions:
        install_rich_traceback()


# IPython extension. Must be imported by top-level module.
def load_ipython_extension(ipython):
    """
    The Eradiate notebook extension.

    This extension simply calls the :func:`eradiate.notebook.install` function.

    See Also
    --------
    :func:`eradiate.notebook.install`

    Notes
    -----
    This extension should be loaded using the IPython magic:

    .. code::

       %load_ext eradiate
    """

    install()



from __future__ import annotations

from abc import ABC, abstractmethod

import attrs

from ..core import CompositeSceneElement, NodeSceneElement
from ..._factory import Factory
from ...attrs import documented, get_doc, parse_docs

surface_factory = Factory()
surface_factory.register_lazy_batch(
    [
        ("_basic.BasicSurface", "basic", {}),
        ("_central_patch.CentralPatchSurface", "central_patch", {}),
        ("_dem.DEMSurface", "dem", {}),
    ],
    cls_prefix="eradiate.scenes.surface",
)


@parse_docs
@attrs.define(eq=False, slots=False)
class Surface(CompositeSceneElement, ABC):
    """
    An abstract base class defining common facilities for all surfaces.

    All scene elements deriving from this interface are composite and cannot be
    turned into a kernel scene instance on their own: they must be owned by a
    container which take care of expanding them.

    Notes
    -----
    * This class is to be used as a mixin.
    """

    id: str | None = documented(
        attrs.field(
            default="surface",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(NodeSceneElement, "id", "doc"),
        type=get_doc(NodeSceneElement, "id", "type"),
        init_type=get_doc(NodeSceneElement, "id", "init_type"),
        default='"surface"',
    )

    @property
    @abstractmethod
    def _template_bsdfs(self) -> dict:
        pass

    @property
    @abstractmethod
    def _template_shapes(self) -> dict:
        pass

    @property
    def template(self) -> dict:
        # Inherit docstring
        return {**self._template_bsdfs, **self._template_shapes}

    @property
    @abstractmethod
    def _params_bsdfs(self) -> dict:
        pass

    @property
    @abstractmethod
    def _params_shapes(self) -> dict:
        pass

    @property
    def params(self) -> dict:
        # Inherit docstring
        return {**self._params_bsdfs, **self._params_shapes}



import typing as t

import xarray as xr

XarrayObj = t.Union[xr.DataArray, xr.Dataset]


def unstack_mdistant_grid(obj: XarrayObj) -> XarrayObj:
    """
    Reindex data produced by a :meth:`.MultiDistantMeasure.grid()` measure on
    (VAA, VZA) dimensions.

    This effectively reshapes the data, which is initially indexed on film width
    and height dimensions (the latter having a single element).

    Parameters
    ----------
    obj : DataArray or Dataset
        An xarray object produced by processing a
        :meth:`.MultiDistantMeasure.grid()` measure, indexed on film dimensions.

    Returns
    -------
    DataArray or Dataset
        An object of the same type, reindexed on VZA and VAA coordinates.
    """
    # Collect metadata
    vza_attrs = obj["vza"].attrs
    vaa_attrs = obj["vaa"].attrs

    # Build new index
    midx_values = (obj.vza.values.flatten(), obj.vaa.values.flatten())
    result = obj.drop_vars(
        ["x_index", "x"]
    )  # Remove the x film coordinate (irrelevant after unstacking)
    result = result.set_xindex(
        ["vza", "vaa"], tuples=midx_values
    )  # Reindex the x_index dimension with a multi-index using the vza and vaa variables
    result = result.unstack()  # Unstack the multi-level index

    # Reapply metadata
    result["vza"].attrs = vza_attrs
    result["vaa"].attrs = vaa_attrs

    return result



from ._astro_object import AstroObjectIllumination as AstroObjectIllumination
from ._constant import ConstantIllumination as ConstantIllumination
from ._core import AbstractDirectionalIllumination as AbstractDirectionalIllumination
from ._core import Illumination as Illumination
from ._core import illumination_factory as illumination_factory
from ._directional import DirectionalIllumination as DirectionalIllumination
from ._spot import SpotIllumination as SpotIllumination



from __future__ import annotations

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr

from ._core import ShapeNode
from ..bsdfs import BSDF
from ..core import BoundingBox
from ... import validators
from ...attrs import documented, parse_docs
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


def _normalize(v: np.typing.ArrayLike) -> np.ndarray:
    # Return L2 norm of a vector
    return np.array(v) / np.linalg.norm(v)


def _edges_converter(value):
    # Basic unit conversion and array reshaping
    length_units = ucc.get("length")
    value = np.reshape(
        pinttr.util.ensure_units(value, default_units=length_units).m_as(length_units),
        (-1,),
    )

    # Broadcast if relevant
    if len(value) == 1:
        value = np.full((2,), value[0])

    return value * length_units


@parse_docs
@attrs.define(eq=False, slots=False)
class RectangleShape(ShapeNode):
    """
    Rectangle shape [``rectangle``].

    This shape represents a rectangle parametrized by the length of its edges,
    the coordinates of its central point, a normal vector and an orientation
    vector.

    Notes
    -----
    * If the `to_world` parameter is set, it will override the other parameters.
      In that case, the transformation will be applied to a rectangle that covers the
      [-1, -1, 0] to [1, 1, 0] area, with a normal pointing towards [0, 0, 1].
    """

    edges: pint.Quantity = documented(
        pinttr.field(
            factory=lambda: [1, 1],
            converter=_edges_converter,
            units=ucc.deferred("length"),
        ),
        doc="Length of the rectangle's edges. "
        "Unit-enabled field (default: ``ucc['length']``).",
        type="quantity",
        init_type="quantity or array-like, optional",
        default="[1, 1]",
    )

    center: pint.Quantity = documented(
        pinttr.field(factory=lambda: [0, 0, 0], units=ucc.deferred("length")),
        doc="Cartesian coordinates of the rectangle's central point. "
        "Unit-enabled field (default: ``ucc['length]``).",
        type="quantity",
        init_type="quantity or array-like, optional",
        default="[0, 0, 0]",
    )

    normal: np.ndarray = documented(
        attrs.field(
            factory=lambda: [0, 0, 1],
            converter=_normalize,
            validator=validators.is_vector3,
        ),
        doc="Normal vector of the plane containing the rectangle. Defaults to "
        "the +Z direction.",
        type="array",
        init_type="array-like, optional",
        default="[0, 0, 1]",
    )

    up: np.ndarray = documented(
        attrs.field(
            factory=lambda: [0, 1, 0],
            converter=_normalize,
            validator=validators.is_vector3,
        ),
        doc="Orientation vector defining the rotation of the rectangle around "
        "the normal vector. Defaults to the +Y direction.",
        type="array",
        init_type="array-like, optional",
        default="[0, 1, 0]",
    )

    def bbox(self) -> BoundingBox:
        # Inherit docstring
        raise NotImplementedError

    @property
    def template(self) -> dict:
        # Inherit docstring
        length_units = uck.get("length")
        scale = self.edges.m_as(length_units) * 0.5
        if self.to_world is not None:
            trafo = self.to_world
        else:
            trafo = mi.ScalarTransform4f.look_at(
                origin=self.center.m_as(length_units),
                target=self.center.m_as(length_units) + self.normal,
                up=self.up,
            ) @ mi.ScalarTransform4f.scale([scale[0], scale[1], 1.0])

        result = {
            "type": "rectangle",
            "to_world": trafo,
        }

        return result

    @classmethod
    def surface(
        cls,
        altitude: pint.Quantity = 0.0 * ureg.km,
        width: pint.Quantity = 1.0 * ureg.km,
        bsdf: BSDF | None = None,
    ) -> RectangleShape:
        """
        This class method constructor provides a simplified parametrization of
        the rectangle shape better suited for the definition of the surface when
        configuring the one-dimensional model.

        The resulting rectangle shape is a square with edge length equal to
        `width`, centred at [0, 0, `altitude`], with normal vector +Z.

        Parameters
        ----------
        altitude : quantity or array-like, optional, default: 0 km
            Surface altitude. If a unitless value is passed, it is interpreted
            as ``ucc['length']``.

        width : quantity or float, optional, default: 1 km
            Edge length. If a unitless value is passed, it is interpreted
            as ``ucc['length']``.

        bsdf : BSDF or dict, optional, default: None
            A BSDF specification, forwarded to the main constructor.

        Returns
        -------
        RectangleShape
            A rectangle shape which can be used as the surface in a
            plane parallel geometry.
        """
        altitude = pinttr.util.ensure_units(altitude, default_units=ucc.get("length"))
        return cls(
            edges=width,
            center=[0.0, 0.0, altitude.m] * altitude.units,
            normal=np.array([0, 0, 1]),
            up=np.array([0, 1, 0]),
            bsdf=bsdf,
        )



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr

from .shapes import CuboidShape, RectangleShape, Shape, SphereShape
from ..attrs import documented, parse_docs
from ..constants import EARTH_RADIUS
from ..kernel import map_cube, map_unit_cube
from ..radprops import ZGrid
from ..units import unit_context_config as ucc
from ..units import unit_context_kernel as uck
from ..units import unit_registry as ureg


@parse_docs
@attrs.define
class SceneGeometry(ABC):
    """
    Abstract base class defining a scene geometry.

    Warnings
    --------
    If a ``zgrid`` value is passed to the constructor (instead of letting
    the constructor set it automatically), its extent must be
    [``ground_altitude``, ``toa_altitude``]. The constructor will raise
    a :class:`ValueError` otherwise.
    """

    toa_altitude: pint.Quantity = documented(
        pinttr.field(default=120.0 * ureg.km, units=ucc.deferred("length")),
        doc="Top-of-atmosphere level altitude. "
        'Unit-enabled field (default: ``ucc["length"]``).',
        default="120 km",
        type="pint.Quantity",
        init_type="float or quantity",
    )

    ground_altitude: pint.Quantity = documented(
        pinttr.field(default=0.0 * ureg.km, units=ucc.deferred("length")),
        doc="Baseline ground altitude. "
        'Unit-enabled field (default: ``ucc["length"]``).',
        default="0 km",
        type="pint.Quantity",
        init_type="float or quantity",
    )

    zgrid: ZGrid = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(
                lambda x: ZGrid(x) if not isinstance(x, ZGrid) else x
            ),
            validator=attrs.validators.optional(attrs.validators.instance_of(ZGrid)),
        ),
        doc="The altitude mesh on which the radiative properties of "
        "heterogeneous atmosphere components are evaluated. "
        "If unset, a default grid with one layer per 100 m (or 10 layers if "
        "the atmosphere object height is less than 100 m) is used.",
        type=".ZGrid",
        init_type=".ZGrid, quantity or ndarray, optional",
    )

    def __attrs_post_init__(self) -> None:
        # Set altitude grid
        if self.zgrid is None:
            bottom = self.ground_altitude.m_as(ureg.m)
            top = self.toa_altitude.m_as(ureg.m)
            step = min(100.0, (top - bottom) / 10.0)
            self.zgrid = ZGrid(
                ureg.convert(
                    np.arange(bottom, top + step * 0.1, step),
                    ureg.m,
                    ucc.get("length"),
                )
            )

        else:
            grid_bottom = self.zgrid.levels[0]
            if not np.isclose(grid_bottom, self.ground_altitude):
                raise ValueError(
                    "zgrid bottom must match ground_altitude; "
                    f"expected {self.ground_altitude}, got {grid_bottom}"
                )

            grid_top = self.zgrid.levels[-1]
            if not np.isclose(grid_top, self.toa_altitude):
                raise ValueError(
                    "zgrid top must match toa_altitude; "
                    f"expected {self.toa_altitude}, got {grid_top}"
                )

    @classmethod
    def convert(cls, value: t.Any) -> t.Any:
        """
        Attempt conversion of a value to a :class:`.SceneGeometry` subtype.

        Parameters
        ----------
        value
            Value to attempt conversion of. If a dictionary is passed, its
            ``"type"`` key is used to route its other entries as keyword
            arguments to the appropriate subtype's constructor. If a string is
            passed, this method calls itself with the parameter
            ``{"type": value}``.

        Returns
        -------
        result
            If `value` is a dictionary, the constructed :class:`.SceneGeometry`
            instance is returned. Otherwise, `value` is returned.

        Raises
        ------
        ValueError
            A dictionary was passed but the requested type is unknown.
        """
        if isinstance(value, str):
            return cls.convert({"type": value})

        if isinstance(value, dict):
            value = value.copy()
            geometry_type = value.pop("type")

            # Note: if this conditional becomes large, use a dictionary
            if geometry_type == "plane_parallel":
                geometry_cls = PlaneParallelGeometry
            elif geometry_type == "spherical_shell":
                geometry_cls = SphericalShellGeometry
            else:
                raise ValueError(f"unknown geometry type '{geometry_type}'")

            return geometry_cls(**pinttr.interpret_units(value, ureg=ureg))

        return value

    @property
    @abstractmethod
    def atmosphere_shape(self) -> Shape:
        """
        :class:`.Shape`: Stencil of the participating medium representing the
        atmosphere.
        """
        pass

    @property
    @abstractmethod
    def atmosphere_volume_to_world(self) -> mi.ScalarTransform4f:
        """
        :class:`mi.ScalarTransform4f` : Mitsuba transform mapping volume texture
            coordinates to world coordinates for heterogeneous atmosphere
            components.
        """
        pass

    @property
    @abstractmethod
    def surface_shape(self) -> Shape:
        """
        :class:`.Shape` : Shape representing the surface.
        """
        pass


@parse_docs
@attrs.define
class PlaneParallelGeometry(SceneGeometry):
    """
    Plane parallel geometry.

    A plane parallel atmosphere is translation-invariant in the X and Y
    directions. However, Eradiate represents it with a finite 3D geometry
    consisting of a cuboid. By default, the cuboid's size is computed
    automatically; however, it can also be forced by assigning a value to
    the `width` field.
    """

    width: pint.Quantity = documented(
        pinttr.field(default=1e6 * ureg.km, units=ucc.deferred("length")),
        doc="Cuboid shape width.",
        type="quantity",
        init_type="quantity or float",
        default="1,000,000 km",
    )

    @property
    def atmosphere_shape(self) -> CuboidShape:
        return CuboidShape.atmosphere(
            top=self.toa_altitude, bottom=0.0 * ureg.km, width=self.width
        )

    @property
    def atmosphere_volume_to_world(self) -> mi.ScalarTransform4f:
        length_units = uck.get("length")
        shape = self.atmosphere_shape

        # The bounding box corresponds to the vertices of the cuboid
        center = shape.center.m_as(length_units)
        edges = shape.edges.m_as(length_units)
        xmin, ymin = center[0:2] - edges[0:2] * 0.5
        xmax, ymax = center[0:2] + edges[0:2] * 0.5
        zmin = self.ground_altitude.m_as(length_units)
        zmax = self.toa_altitude.m_as(length_units)

        return map_unit_cube(xmin, xmax, ymin, ymax, zmin, zmax)

    @property
    def surface_shape(self) -> RectangleShape:
        return RectangleShape.surface(altitude=self.ground_altitude, width=self.width)


@parse_docs
@attrs.define
class SphericalShellGeometry(SceneGeometry):
    """
    Spherical shell geometry.

    A spherical shell atmosphere has a spherical symmetry. Eradiate represents
    it with a finite 3D geometry consisting of a sphere. By default, the
    sphere's radius is set equal to Earth's radius.
    """

    planet_radius: pint.Quantity = documented(
        pinttr.field(default=EARTH_RADIUS, units=ucc.deferred("length")),
        doc="Planet radius. Defaults to Earth's radius.",
        type="quantity",
        init_type="quantity or float",
        default=":data:`.EARTH_RADIUS`",
    )

    @property
    def atmosphere_shape(self) -> SphereShape:
        return SphereShape.atmosphere(
            top=self.toa_altitude, planet_radius=self.planet_radius
        )

    @property
    def atmosphere_volume_to_world(self) -> mi.ScalarTransform4f:
        length_units = ucc.get("length")

        # The bounding box corresponds to the vertices of the bounding box
        bbox = self.atmosphere_shape.bbox
        xmin, ymin, zmin = bbox.min.m_as(length_units)
        xmax, ymax, zmax = bbox.max.m_as(length_units)

        return map_cube(xmin, xmax, ymin, ymax, zmin, zmax)

    @property
    def atmosphere_volume_rmin(self) -> float:
        """
        Value for the ``rmin`` parameter of the ``sphericalcoordsvolume`` plugin
        describing the volume data in the spherical shell geometry.
        """
        return (self.planet_radius / (self.planet_radius + self.toa_altitude)).m_as(
            ureg.dimensionless
        )

    @property
    def surface_shape(self) -> Shape:
        return SphereShape.surface(
            altitude=self.ground_altitude, planet_radius=self.planet_radius
        )



from __future__ import annotations

import datetime
import importlib.util
import typing as t
import warnings

import attrs
import numpy as np
import pint
import xarray as xr

from ._core import Spectrum
from ... import converters, data, validators
from ...attrs import documented, parse_docs
from ...kernel import InitParameter, UpdateParameter
from ...units import PhysicalQuantity, to_quantity
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ...util.misc import summary_repr


def _datetime_converter(x: t.Any):
    if x is not None:
        try:
            import dateutil
        except ModuleNotFoundError:
            warnings.warn(
                "To use the date-based Solar irradiance scaling feature, you "
                "must install dateutil.\n"
                "See instructions on https://dateutil.readthedocs.io/#installation."
            )
            raise

        try:
            importlib.import_module("astropy")
        except ModuleNotFoundError:
            warnings.warn(
                "To use the date-based Solar irradiance scaling feature, you "
                "must install astropy.\n"
                "See instructions on https://www.astropy.org/."
            )
            raise

        return dateutil.parser.parse(x)


@parse_docs
@attrs.define(eq=False, slots=False)
class SolarIrradianceSpectrum(Spectrum):
    """
    Solar irradiance spectrum [``solar_irradiance``].

    This scene element produces the scene dictionary required to
    instantiate a kernel plugin using the Sun irradiance spectrum. The data set
    used by this element is controlled by the ``dataset`` field.

    See Also
    --------

    :ref:`Solar irradiance spectrum data guide <sec-data-solar_irradiance>`

    Notes
    ------

    * The spectral range of the data sets shipped can vary and an attempt for
      use outside the supported spectral range will raise a
      :class:`ValueError` upon calling :meth:`kernel_dict`.

    * When the ``datetime`` field is set, the spectrum is automatically scaled
      to account for the seasonal variations of the Earth-Sun distance using the
      ephemeris of :func:`astropy.coordinates.get_sun`.
      The dataset is assumed to be normalized to an Earth-Sun distance of 1 AU.
      This will trigger the import of :mod:`astropy.coordinates` and consume a
      significant amount of memory (150 MiB with astropy v5.1).

    * The ``scale`` field can be used to apply additional arbitrary scaling.
      It is mostly used for debugging purposes. It can also be used to rescale
      user-defined spectra normalized at an Earth-Sun distance different from
      1 AU.

    * The evaluation method depends on the active mode:

      * in ``mono_*`` modes, the spectrum is evaluated at the spectral context
        wavelength;
      * in ``ckd_*`` modes, the spectrum is evaluated as the average value over
        the spectral context bin (the integral is computed using a trapezoid
        rule).

    * The produced kernel dictionary automatically adjusts its irradiance units
      depending on the selected kernel default units.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    quantity: PhysicalQuantity = attrs.field(
        default=PhysicalQuantity.IRRADIANCE, init=False, repr=False
    )

    dataset: xr.Dataset = documented(
        attrs.field(
            default="coddington_2021-1_nm",
            converter=converters.to_dataset(
                load_from_id=lambda x: data.load_dataset(
                    f"spectra/solar_irradiance/{x}.nc",
                )
            ),
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Solar irradiance spectrum dataset. "
        "If a xarray.Dataset is passed, the dataset is used as is "
        "(refer to the data guide for the format requirements of this dataset)."
        "If a path is passed, the converter tries to open the corresponding "
        "file on the hard drive; should that fail, it queries the Eradiate data"
        "store with that path."
        "If a string is passed, it is interpreted as a Solar irradiance "
        "spectrum identifier "
        "(see :ref:`sec-data-solar_irradiance` for the list); ",
        type="Dataset",
        init_type="Dataset or str or path-like, optional",
        default='"coddington_2021-1_nm"',
    )

    scale: float = documented(
        attrs.field(default=1.0, converter=float, validator=validators.is_positive),
        doc="Arbitrary scaling factor. This scaling factor is applied in "
        "addition to the datetime-based scaling controlled by the *datetime* "
        "parameter.",
        type="float or datetime",
        init_type="float or datetime or str",
        default="1.0",
    )

    datetime: datetime.datetime | None = documented(
        attrs.field(
            default=None,
            converter=_datetime_converter,
        ),
        type="datetime or None",
        init_type="datetime or str, optional",
        doc="Date for which the spectrum is to be evaluated. An ISO "
        "string can be passed and will be interpreted by "
        ":meth:`dateutil.parser.parse`. This parameter scales the irradiance "
        "spectrum to account for the seasonal variation of the Earth-Sun "
        "distance. This scaling is applied in addition to the arbitrary "
        "scaling controlled by the *scale* parameter.",
    )

    def _scale_earth_sun_distance(self) -> float:
        """
        Compute scaling factor applied to the irradiance spectrum based on the
        Earth-Sun distance.
        """
        # Note: We assume that the loaded dataset is for a reference
        # Earth-Sun distance of 1 AU
        if self.datetime is None:
            return 1.0

        else:
            # Note: astropy.coordinates consumes a significant amount of memory
            # (150 MiB with astropy v5.1). The import is therefore optional for
            # performance.
            import astropy.coordinates
            import astropy.time
            import astropy.units

            # The irradiance scales as the inverse of d**2, where d is the
            # Earth-Sun distance divided by the AU (reference distance for all
            # Solar irradiance spectra in Eradiate).
            return (
                float(
                    astropy.units.au
                    / astropy.coordinates.get_sun(
                        astropy.time.Time(self.datetime)
                    ).distance
                )
                ** 2
            )

    def eval_mono(self, w: pint.Quantity) -> pint.Quantity:
        # Inherit docstring

        w_units = ureg(self.dataset.ssi.w.attrs["units"])
        irradiance = to_quantity(
            self.dataset.ssi.interp(w=w.m_as(w_units), method="linear")
        )

        # Raise if out of bounds or ill-formed dataset
        if np.any(np.isnan(irradiance.magnitude)):
            raise ValueError("interpolation of solar irradiance dataset returned nan")

        return irradiance * self.scale * self._scale_earth_sun_distance()

    def eval_ckd(self, w: pint.Quantity, g: float) -> pint.Quantity:
        return self.eval_mono(w=w)

    def integral(self, wmin: pint.Quantity, wmax: pint.Quantity) -> pint.Quantity:
        raise NotImplementedError

    @property
    def template(self) -> dict:
        # Inherit docstring

        return {
            "type": "uniform",
            "value": InitParameter(
                evaluator=lambda ctx: float(
                    self.eval(ctx.si).m_as(uck.get("irradiance"))
                )
            ),
        }

    @property
    def params(self) -> dict:
        # Inherit docstring

        return {
            "value": UpdateParameter(
                evaluator=lambda ctx: float(
                    self.eval(ctx.si).m_as(uck.get("irradiance"))
                ),
                flags=UpdateParameter.Flags.SPECTRAL,
            )
        }



"""
Logging facilities.
"""

import logging
import re

import mitsuba.scalar_rgb as mi
from tqdm.auto import tqdm

from .. import config


def _add_logging_level(level_name, level_num, method_name=None):
    """
    Comprehensively adds a new logging level to the ``logging`` module and the
    currently configured logging class.

    ``level_name`` becomes an attribute of the ``logging`` module with the value
    ``level_num``. ``method_name`` becomes a convenience method for both
    ``logging`` itself and the class returned by ``logging.getLoggerClass()``
    (usually just ``logging.Logger``). If ``method_name`` is not specified,
    ``level_name.lower()`` is used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an ``AttributeError`` if the level name is already an attribute of the
    ``logging`` module or if the method name is already present

    Example
    -------
    >>> _add_logging_level('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5
    """
    # See also: https://stackoverflow.com/a/35804945/3645374
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        raise AttributeError(f"{level_name} already defined in logging module")
    if hasattr(logging, method_name):
        raise AttributeError(f"{method_name} already defined in logging module")
    if hasattr(logging.getLoggerClass(), method_name):
        raise AttributeError(f"{method_name} already defined in logger class")

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


# Regex for log parsing
mi_log_parser = re.compile(
    r"^(?P<datetime>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})\s*"
    r"(?P<level>\S*)\s*"
    r"(?P<thread>\S*)\s*"
    r"\[(?P<class>\S*)\]\s*"
    r"(?P<message>.*)$"
)

# Mitsuba Logger instance
mi_logger = None

# Module-local logger
logger = logging.getLogger("mitsuba")


def _get_logger():
    """
    Get Mitsuba Logger instance
    """
    return mi.Thread.thread().logger()


def install_logging(force: bool = False) -> None:
    """
    Activate Python logging framework for Mitsuba.

    Parameters
    ----------
    force : bool, optional
        If ``True``, force-reactivate logging.
    """
    global mi_logger

    if mi_logger is not None and not force:
        return

    try:
        _add_logging_level("TRACE", logging.DEBUG - 5, "trace")
    except AttributeError:
        pass

    LOGLEVEL_DISPATCH = {
        mi.LogLevel.Trace: "trace",
        mi.LogLevel.Debug: "debug",
        mi.LogLevel.Info: "info",
        mi.LogLevel.Warn: "warning",
        mi.LogLevel.Error: "error",
    }

    class EradiateAppender(mi.Appender):
        def __init__(self):
            super().__init__()
            self.reset()

        def reset(self):
            self.bar = None
            self.progress = None

        def append(self, level, text):
            m = mi_log_parser.match(text)
            msg = (
                f"{m.group('thread')} [{m.group('class')}] {m.group('message')}"
                if m is not None
                else text
            )
            getattr(logger, LOGLEVEL_DISPATCH[level])(msg)

        def log_progress(self, progress, name, formatted, eta, ptr=None):
            if self.bar is None:
                self.bar = tqdm(
                    desc="Mitsuba",
                    initial=0.0,
                    total=1.0,
                    unit_scale=1.0,
                    leave=True,
                    bar_format="{l_bar}{bar}| {elapsed}, ETA={remaining}",
                    disable=config.settings.progress < config.ProgressLevel.KERNEL,
                )
                self.progress = 0.0

            self.bar.update(progress - self.progress)
            self.progress = progress

            if progress >= 1.0:
                self.bar.close()
                self.reset()

    if mi_logger is None:
        mi_logger = _get_logger()
        mi_logger.clear_appenders()
        mi_logger.add_appender(EradiateAppender())
        mi_logger.set_log_level(mi.LogLevel.Info)



from ._buffermesh import BufferMeshShape as BufferMeshShape
from ._core import Shape as Shape
from ._core import ShapeInstance as ShapeInstance
from ._core import ShapeNode as ShapeNode
from ._core import shape_factory as shape_factory
from ._cuboid import CuboidShape as CuboidShape
from ._filemesh import FileMeshShape as FileMeshShape
from ._rectangle import RectangleShape as RectangleShape
from ._sphere import SphereShape as SphereShape
